"""A module for post processing BigDFT calculations.

"""


def _system_command(command, options):
    """
    Run the command as ``os.system (command + options)``

    Args:
       command (str): the actual command to run.
       options (str): the options to pass to the command.
    """
    from subprocess import call

    command_str = command + " " + options
    call(command_str, shell=True)


def _get_datadir(log):
    """
    Given a log file, this returns the path to the data directory.

    Args:
      log (Logfile): logfile from a BigDFT run.
    Returns:
      str: path to the data directory.
    """
    from os.path import join
    return join(log.srcdir, log.data_directory)


def _get_target(pv, cutoff, rand):
    from numpy.random import choice
    # Find the least pure fragment to modify
    target = min(pv, key=pv.get)
    target_purity = pv[target]

    if abs(target_purity) >= cutoff and rand:
        target_list = [pid for pid in pv if abs(pv[pid]) > cutoff]
        plist = [abs(x) for x in pv.values() if abs(x) > cutoff]
        normalize = sum(plist)
        plist = [x/normalize for x in plist]
        target = choice(target_list, p=plist)
    return target, target_purity


def _get_distance_dict(tree, frag_lookup, target, target_frag, distcut, k):
    # Build a list of nearby atoms for consideration
    ndist, nearest = tree.query(target_frag, distance_upper_bound=distcut, k=k)
    ndist = [x for sub in ndist for x in sub]
    nearest = [x for sub in nearest for x in sub]

    # Update that into a list of nearby fragments, and keep track of
    # the closest fragment
    distdict = {}
    for idx, dist in zip(nearest, ndist):
        try:
            fragidx = frag_lookup[idx]
        except IndexError:
            # kdtree returns invalid indices if it can't find enough
            # points.
            continue
        if fragidx == target:
            continue
        if fragidx not in distdict:
            distdict[fragidx] = dist
        elif distdict[fragidx] > dist:
            distdict[fragidx] = dist
    return distdict


class BigDFTool(object):
    """

    This module defines the actions that can be performed using the
    ``bigdft-tool`` python script. Such a script sometimes invokes the
    ``memguess`` executable, which is executed serially. Sometimes the
    ``utilities`` main program is necessary to take care of more time
    consuming operations which require parallelism. We might also use
    the ``obabel`` wrapper provided by BigDFT.

    Args:
      omp (int): number of OpenMP threads.
        It defaults to the $OMP_NUM_THREADS variable in the environment,
        if present, otherwise it fixes the run to 1 thread.
      mpi_run (str): define the MPI command to be used.
        It defaults to the value $BIGDFT_MPIRUN of the environment, if present.
        When using this calculator into a job submission script, the value of
        $BIGDFT_MPIRUN variable may be set appropriately to launch the bigdft
        executable.
    """

    def __init__(self, omp=1, mpi_run=None):
        from futile.YamlArgparse import get_python_function
        from futile import YamlIO
        from os import environ, pathsep
        from os.path import join
        from functools import partial
        from copy import deepcopy

        # Executables
        self.bigdftroot = environ.get('BIGDFT_ROOT')
        # environ['OMP_NUM_THREADS'] = str(omp)
        if not mpi_run:
            mpi_run = environ.get("BIGDFT_MPIRUN")

        # Load the dictionary that defines all the operations.
        # This path should be easier to specify if you use the python
        # egg setup.
        found = False
        for path in environ.get("PYTHONPATH", '').split(pathsep):
            try:
                with open(join(path, "BigDFT", "Database",
                               "postprocess.yaml")) as ifile:
                    db = YamlIO.load(stream=ifile)[0]
                    found = True
            except IOError:
                continue
        if not found and self.bigdftroot is not None:
            raise IOError("Couldn't find postprocess.yaml. " +
                          "Check your python path.")

        # Create the subroutines of the class from the dictionary
        if self.bigdftroot is not None:
            for action, spec in db.items():
                naction = action.replace("_", "-")
                nspec = deepcopy(spec)
                cmd = join(self.bigdftroot, nspec["tool"])
                if nspec["mpi"]:
                    cmd = mpi_run + " " + cmd
                else:
                    nspec["args"]["mpirun"] = {"default": "" + mpi_run + ""}
                    nspec["args"]["action"] = {"default": "" + naction + ""}
                    my_action = get_python_function(
                        partial(self._invoke_command, cmd), action, nspec)
                    setattr(self, action, my_action)

    def _auto_fragment(self, msys, cutoff, verbose, rand, view, criteria):
        from scipy.spatial import KDTree

        # Setup one big kdtree and reverse lookup.
        poslist = []
        posdict = {}
        frag_lookup = []
        # Note the iteration is not over msys because the view may be a subset.
        for fragid in view.purities:
            posdict[fragid] = []
            for at in msys[fragid]:
                pos = at.get_position()
                poslist.append(pos)
                posdict[fragid].append(pos)
                frag_lookup.append(fragid)
        tree = KDTree(poslist)

        fragment_mapping = {frag: [] for frag in view.purities}
        starting_length = len(fragment_mapping)

        # Cutoff for how far to look for a fragment to add.
        distcut = 10.0
        k = 10

        # Zero the internal bond order so we never merge a fragment with itself
        for fragid in view.bond_orders:
            view.bond_orders[fragid][fragid] = 0

        for i in range(0, starting_length - 1):
            if verbose and starting_length > 20:
                step = int(starting_length / 20.0)
                if i % step == 0:
                    print(str(5 * (i / step)) + "% done",
                          len(fragment_mapping))
            target, target_purity = _get_target(view.purities, cutoff, rand)
            if abs(target_purity) < cutoff:
                break

            distdict = _get_distance_dict(tree, frag_lookup, target,
                                          posdict[target], distcut, k)

            # In case nothing was found, we widen the search distance.
            if len(distdict.keys()) == 0:
                distcut *= 2
                k *= 2
                continue

            # Find the neighbor to add in
            if criteria == 'distance':
                addition = min(distdict, key=distdict.get)
            elif criteria == 'bondorder':
                bo_target = {x: view.bond_orders[target][x] for x in distdict}
                addition = max(bo_target, key=bo_target.get)
            else:
                raise ValueError("Criteria must be distance or bondorder")

            # New purity values.
            new_purity = view.bond_orders[target][addition] + \
                view.bond_orders[addition][target]
            new_purity += view.purities[target]*view.charges[target]/2.0
            new_purity += view.purities[addition]*view.charges[addition]/2.0
            new_purity *= 2.0 / (view.charges[target] + view.charges[addition])
            view.purities[target] = new_purity

            posdict[target] = posdict[target] + posdict[addition]

            # New bond order.
            for fragid in view.bond_orders:
                view.bond_orders[target][fragid] += \
                    view.bond_orders[addition][fragid]
                view.bond_orders[fragid][target] += \
                    view.bond_orders[fragid][addition]
            view.bond_orders[target][target] = 0

            # New charge.
            view.charges[target] += view.charges[addition]

            # update lookups
            for fragid in list(view.bond_orders):
                del view.bond_orders[fragid][addition]
            del view.bond_orders[addition]
            del posdict[addition]
            del view.purities[addition]
            del view.charges[addition]
            fragment_mapping[target] += [addition] + fragment_mapping[addition]
            del fragment_mapping[addition]

            # Update the fragment lookup
            frag_lookup = [target if x == addition else x for x in frag_lookup]

        return fragment_mapping

    def auto_fragment(self, system, view, cutoff, verbose=False,
                      rand=False, criteria="bondorder"):
        """
        Refragment a system based on the results of a previous calculation.

        The auto fragment protocol performs a greedy optimization using either
        the bond order or the distance between fragments as a guide. The
        purity values and the fragment bond orders are essential quantities
        for this calculation, so they can be passed if they are already
        cached.

        By using the `rand` keyword, you can trigger a stochastic
        refragmentation process. If this process is repeated many times, you
        may find a result that improves upon the greedy optimization approach.

        Args:
          system (BigDFT.Systems.System): the system to fragment.
          view (BigDFT.Systems.System): a view of the system.
          cutoff (float): the termination criteria. When the worst fragment is
            more pure than this cutoff, the calculation stops.
          verbose (bool): whether to print out as it proceeds.
          rand (bool): whether to activate the stochastic approach.
          criteria (string): either distance or bondorder.

        Returns:
          dict: a mapping from the old system to the new system where each
          fragment fullfills the purity criteria.
        """
        from copy import deepcopy

        # This deep copy is manually overrided for a view so this is not
        # too slow.
        new_view = deepcopy(view)

        # Auto Fragment.
        fragment_mapping = self._auto_fragment(system, cutoff, verbose,
                                               rand, new_view, criteria)

        # Rename
        remapping = {}
        for fragid, add_list in fragment_mapping.items():
            new_name = "+".join([fragid] + add_list)
            remapping[new_name] = [fragid] + add_list

        return remapping

    def fragment_mask_matrix(self, sys, mat, fragments, log):
        """
        Sometimes we don't want to store an entire matrix, just the parts
        related to some fragments of interest. This routine will mask out a
        matrix, keeping entries only related to the list of fragments
        provided.

        Args:
          sys (BigDFT.Systems.System): the system associated with the matrix.
          mat (scipy.sparse.csr_matrix): the matrix to mask.
          fragments (list): a list of fragment ids to keep.
          log (BigDFT.Logfiles.Logfile): the logfile associated with this
            matrix's calculation.
        Returns:
          (scipy.sparse.csr_matrix): the masked matrix.
        """
        from BigDFT.Systems import System
        from copy import deepcopy
        from scipy.sparse import dok_matrix

        subsys = System()
        for fragid in fragments:
            subsys[fragid] = deepcopy(sys[fragid])

        frag_indices = self.get_frag_indices(subsys, log)
        frag_indices = sum(frag_indices.values(), [])

        imat = dok_matrix(mat.shape)
        for i in frag_indices:
            imat[i, i] = 1

        return imat.dot(mat).dot(imat)

    def compute_fragment_dos(self, frag, log, ks_coeff, eigvals,
                             frag_indices=None, smat=None, assume_pure=False,
                             **kwargs):
        """
        Compute the partial density of states projected onto a given
        fragment.

        Args:
            sys (BigDFT.Fragments.Fragment): the fragment to project on to.
            log (BigDFT.Logfiles.Logfile): the log of the calculation.
            ks_coeff (scipy.sparse.csc_matrix): the matrix of eigenvectors.
            eigvals (list): a list of eigenvalues.
            frag_indices (list): list of indices associated with this
              fragment.
            smat (scipy.sparse.csc_matrix): the overlap matrix.
            assume_pure (bool): an optimization can be performed if we assume
              the target is pure.
            kwargs (dict): any extra options to pass to the DoS constructor.
        Returns:
          (BigDFT.DoS.DoS): a density of states object built using the partial
          density of states.
        """
        from BigDFT.Systems import System
        from BigDFT.DoS import DoS
        # Subsystem with just this fragment.
        subsys = System()
        subsys["FRAG:0"] = frag

        # Optional Argments
        if smat is None:
            smat = self.get_matrix_s(log)
        if frag_indices is None:
            frag_indices = self.get_frag_indices(subsys, log)["FRAG:0"]

        subsmat = smat[frag_indices, :]
        subkmat = ks_coeff[frag_indices, :].T

        vals = []
        for n in range(len(eigvals)):
            work = subkmat[n, :].dot(subsmat)
            if assume_pure:
                sumval = work[:, frag_indices].dot(ks_coeff[frag_indices, n])
            else:
                sumval = work[:, :].dot(ks_coeff[:, n])
            vals.append(sumval[0, 0])

        # Construct the dos object
        dos = DoS(energies=eigvals, norm=[vals], units='AU', **kwargs)
        return dos

    def create_layered_qmmm_system(self, system, target, pairwise_bo, cutoffs,
                                   criteria="bondorder", link_atoms=False):
        """
        Creates a multilayered system suitable for QM/MM calculations.
        For each layer, a suitable QM region is built around it.

        Args:
          system (BigDFT.Systems.System): a System class, already broken up
            into fragments.
          pairwise_bo (dict): pairwise bond order values between all fragments
            in the system.
          target (str): the name of the fragment to treat as the target of
            the qm/mm run.
          cutoffs (list): a list of cutoff value for fragment interactions.
            The number of layers is equal to the number of cutoffs.
          criteria (str): how to determine which atoms are included in the
            QM region. Valid choices are "bondorder" and "distance".
          link_atoms (bool): whether to generate link atoms.

        Returns:
          (list): a list of Systems, one for each QM layer.
          (System): the MM region.

        """
        from copy import deepcopy
        from BigDFT.Systems import System

        # First, we will build the layers using only keys.
        qm_list = []
        qm, mm = self.create_qmmm_system(system, target, pairwise_bo[target],
                                         cutoffs[0], criteria,
                                         link_atoms=False)
        qm_list.append(list(qm.keys()))
        mm_keys = list(mm.keys())

        # To avoid duplicates
        used_list = deepcopy(list(qm.keys()))

        for k in range(1, len(cutoffs)):
            qm_list.append([])
            for subtarget in qm_list[k-1]:
                qm, mm = self.create_qmmm_system(system,
                                                 subtarget,
                                                 pairwise_bo[subtarget],
                                                 cutoffs[k], criteria,
                                                 link_atoms=False)
                qm_list[k] += qm.keys()

            # Remove duplicates
            qm_list[k] = [x for x in qm_list[k] if x not in used_list]
            qm_list[k] = list(set(qm_list[k]))
            used_list = list(set(used_list + qm_list[k]))

            mm_keys = [x for x in mm_keys if x not in qm_list[k]]

        # Next, we will convert those keys into systems.
        qmsys = []
        for j in range(0, len(cutoffs)):
            qmsys.append(System())
            for k in qm_list[j]:
                qmsys[j][k] = deepcopy(system[k])
        mmsys = System()
        for k in mm_keys:
            mmsys[k] = deepcopy(system[k])

        # Add the link atoms.
        if link_atoms:
            # One big system to add the link atoms.
            mergesys = System()
            for k in range(0, len(cutoffs)):
                for fragid, frag in qmsys[k].items():
                    mergesys[fragid] = frag
            linksub, remove = self.generate_link_atoms(system, mergesys)

            # Insert the link atoms into the correct layers.
            for fragid, frag in linksub.items():
                for at in frag:
                    if at.is_link:
                        for k in range(0, len(cutoffs)):
                            if fragid in qmsys[k]:
                                qmsys[k][fragid] += [at]
                                break

        return qmsys, mmsys

    def create_qmmm_system(self, system, target, bond_order, cutoff,
                           criteria="bondorder", link_atoms=False):
        """
        Creates a system suitable for QM/MM calculations.

        Args:
          system (System): a System class, already broken up into fragments.
          target (str): the name of the fragment to treat as the target of
            the qm/mm run.
          bond_order (dict): bond order values between the target fragment and
            all other fragments in the system.
          cutoff (float): a cutoff value for fragment interactions.
          criteria (str): how to determine which atoms are included in the
            QM region. Valid choices are "bondorder" and "distance".
          link_atoms (bool): whether to generate link atoms.

        Returns:
          (System): the QM region.
          (System): the MM region.
        """
        from BigDFT.Systems import System
        from BigDFT.Fragments import pairwise_distance
        from copy import deepcopy

        qmsys = System()
        mmsys = deepcopy(system)

        if criteria == "bondorder":
            # Order the fragments by their spillage value
            sort_spill = sorted(bond_order.keys(),
                                key=lambda x: bond_order[x], reverse=True)

            # Iterate until we reach the cutoff
            cumsum = sum(bond_order.values())
            for key in sort_spill:
                qmsys[key] = deepcopy(system[key])
                del mmsys[key]

                # check cutoff
                cumsum -= bond_order[key]
                if cumsum < cutoff:
                    break

        elif criteria == "distance":
            interaction = {x: pairwise_distance(system[target], system[x])
                           for x in system.keys()}
            for fragid, frag in system.items():
                if interaction[fragid] < cutoff:
                    qmsys[fragid] = deepcopy(frag)
                    del mmsys[fragid]

        else:
            raise ValueError("Criteria must be either distance or bondorder")

        # Double check that the target is in qmsys.
        # (this can happen if the target is not a pure fragment)
        if target not in qmsys:
            qmsys[target] = system[target]

        if link_atoms:
            qmsys, remove = self.generate_link_atoms(system, qmsys)

        return qmsys, mmsys

    def fragment_bond_order(self, sys, fraglist1, fraglist2, log, kxs=None,
                            frag_indices=None):
        """
        Computes "bond order" between two sets of fragments using the method of
        Mayer. For two atomic fragments, this would describe the bond
        multiplicity of the covalent bond.

        Args:
          sys (BigDFT.Systems.System): the system containing the fragments
            of interest
          fraglist1 (list): a list of fragments to compute the bond order
            of.
          fraglist2 (list): a list of fragments to compute the bond order
            between.
          log (BigDFT.Logfiles.Logfile): the log describing a calculation.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.
          frag_indices (dict): the matrix indices associated with each
            fragment.

        Return:
          (dict): a dictionary of dictionaries, mapping the bond order of each
          fragment in fraglist1 to each fragment in fraglist2.
        """
        from scipy.sparse import csr_matrix, csc_matrix
        # Get the metadata
        if kxs is None:
            kxs = self.get_matrix_kxs(log)

        if frag_indices is None:
            frag_indices = self.get_frag_indices(sys, log)

        slist1 = sorted(fraglist1, key=lambda x: min(frag_indices[x]))
        slist2 = sorted(fraglist2, key=lambda x: min(frag_indices[x]))
        bond_orders = units_quadratic_traces(slist1, slist2, frag_indices,
                                             csc_matrix(kxs),
                                             csr_matrix(kxs), 0.5, 0.5)

        return bond_orders

    def fragment_interaction_energy(self, sys, fraglist1, fraglist2, log,
                                    frag_indices=None, sinvh=None, kxs=None):
        """
        Compute the interaction energies between two sets of fragments.

        Args:
          fraglist1 (list): a list of fragments to compute the interaction
            energy of.
          fraglist2 (list): a list of fragments to compute the interaction
            energy between.
          log (BigDFT.Logfiles.Logfile): the log describing a calculation.
          frag_indices (dict): the matrix indices associated with each
            fragment.
          sinvh (scipy.sparse.csc_matrix): the matrix S^{-1}*H, which might be
            already computed to reduce I/O time.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.

        Returns:
          (float): the projected energy.
        """
        from scipy.sparse import csr_matrix, csc_matrix
        # Get the metadata
        if kxs is None:
            kxs = self.get_matrix_kxs(log)
        if sinvh is None:
            sinvh = self.get_matrix_sinvh(log)

        if frag_indices is None:
            frag_indices = self.get_frag_indices(sys, log)

        interaction_energy = units_quadratic_traces(fraglist1, fraglist2,
                                                    frag_indices,
                                                    csc_matrix(sinvh),
                                                    csr_matrix(kxs))
        return interaction_energy

    def fragment_population(self, sys, fragid, log, kxs=None):
        """
        Performs Mulliken population analysis on a fragment, in case charges
        haven't been computed by doing a multipole analysis.

        Args:
          sys (BigDFT.Systems.System): the system containing the fragment
            of interest
          frag (str): the fragment to compute the charge of.
          log (BigDFT.Logfiles.Logfile): the log describing a calculation.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.

        Return:
          (float): the amount of charge on a fragment.
        """
        from numpy import trace
        from BigDFT.Systems import System

        if kxs is None:
            kxs = self.get_matrix_kxs(log)

        # Get the metadata. It's faster if we perform this operation on
        # a system with just the fragment we're interested in.
        subsys = System()
        subsys[fragid] = sys[fragid]
        frag_indices = self.get_frag_indices(subsys, log)
        indices = frag_indices[fragid]

        smat = kxs[:, indices]
        smat = smat[indices, :]

        return sum(x.nel for x in sys[fragid]) - trace(smat.todense())

    def generate_link_atoms(self, fullsys, subsys, distcut=6.0):
        """
        This routine adds link atoms to a subsystem based on the bond
        order of a full system. Link atom positions are automatically adjusted
        based on the length of some standard bonds.

        Args:
          fullsys (BigDFT.Systems.System): the full system that the subsystem
            is embedded into.
          subsys (BigDFT.Systems.System): the embedded system which needs
            link atoms.
          distcut (float): this cutoff is the largest distance value we expect
            allow a bond to be.

        Returns:
          (BigDFT.Systems.System): the subsystem with link atoms added.
          (BigDFT.Systems.System): a system which has the atoms that were
            removed and replaced with link atoms.
        """
        from BigDFT.Systems import System
        from BigDFT.Fragments import Fragment
        from copy import deepcopy
        from numpy.linalg import norm
        from numpy import array
        from warnings import warn

        # Bond lengths for link atoms.
        bond_lengths = {"C": 2.0598, "N": 1.90862, "O": 1.81414, "F": 1.73855,
                        "P": 2.68341, "S": 2.53223, "Cl": 2.39995,
                        "Br": 2.66451, "I": 3.04246}

        if fullsys.conmat is None:
            raise ValueError("Generating link atoms requires connectivity"
                             " information")

        linksys = deepcopy(subsys)
        removesys = System()

        # Loop over atoms and look for bonds running out of the QM system.
        for fragid in subsys:
            linklist = []
            for i in range(0, len(fullsys[fragid])):
                for ft, bv in fullsys.conmat[fragid][i].items():
                    if ft[0] in subsys.keys():
                        continue
                    if bv == 1:
                        newat = deepcopy(fullsys[ft[0]][ft[1]])
                        # Change the symbol to hydrogen
                        newat.sym = "H"
                        newat.is_link = True
                        # If possible we adjust the position for a reasonable
                        # bond length
                        conat = fullsys[fragid][i]
                        if conat.sym in bond_lengths:
                            pos1 = array(conat.get_position("bohr"))
                            pos2 = array(newat.get_position("bohr"))
                            vec = pos2 - pos1
                            vec *= bond_lengths[conat.sym]/norm(vec)
                            newpos = [x + y for x, y in zip(pos1, vec)]
                            newat.set_position(newpos, units="bohr")
                        else:
                            warn(conat.sym+"bondlength unknown", UserWarning)
                        linklist.append(newat)
                    elif bv > 1:
                        print("Not yet implemented double/triple bonds.", bv)
                        raise NotImplementedError
                    if ft[0] not in removesys:
                        removesys[ft[0]] = Fragment()
                    removesys[ft[0]] += [fullsys[ft[0]][ft[1]]]

            # Add those atoms to the fragment
            for link in linklist:
                linksys[fragid] += [link]

        return linksys, removesys

    def run_compute_purity(self, system, log, kxs=None, frag_list=None,
                           frag_indices=None):
        """
        Compute the purity values of the different system fragments.

        Note that this can also be computed using the fragment multipoles,
        but this provides an implementation for when you don't need those
        values.

        Args:
          system (System): instance of a System class, which defines the
            fragmentation.
          log (Logfile): logfile from the run computed on this system.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.
          frag_list (list): we can also only compute the purity values of some
            select fragments.
          frag_indices (list): the indices of the matrix associated with each
            fragment. This can be precomputed and passed.

        Returns:
          (dict): for each fragment id, what is the purity value.
        """
        from scipy.sparse import csc_matrix

        # Handle the optional parameters
        if frag_list is None:
            working_sys = system
        else:
            working_sys = {k: system[k] for k in frag_list}
        if kxs is None:
            kxs = self.get_matrix_kxs(log)

        # Get the metadata
        if frag_indices is None:
            fidx = self.get_frag_indices(working_sys, log)
        else:
            fidx = {x: frag_indices[x] for x in working_sys}

        # Compute the purity array
        charges = {}
        for fragid, frag in working_sys.items():
            charges[fragid] = sum(x.nel for x in frag)
        purity_array = _compute_purity_values(csc_matrix(kxs), charges, fidx)

        return purity_array

    def get_frag_indices(self, sys, log):
        """
        Compute a lookup table of matrix indices for each fragment in a
        system.

        Args:
          system (System): instance of a System class, which defines the
            fragmentation.
          log (Logfile): logfile from the run computed on this system.

        Returns:
          (dict): a mapping of fragment ids to lists of indices
        """
        from BigDFT.Spillage import MatrixMetadata
        from os.path import join

        data_dir = _get_datadir(log)
        metadatafile = join(data_dir, "sparsematrix_metadata.dat")
        metadata = MatrixMetadata(metadatafile)
        frag_indices = metadata.get_frag_indices(sys)

        return frag_indices

    def get_ks_coeff(self, log):
        """
        Retrieve the Kohn-Sham coefficients and the eigenvalues in matrix
        form.

        log (Logfile): instance of a Logfile class

        Returns:
          (scipy.sparse.csc_matrix): the matrix of coefficients
          (list): list of eigenvalues.
        """
        from os.path import join
        from scipy.sparse import coo_matrix, csc_matrix

        # Define the input files.
        data_dir = _get_datadir(log)
        infile = join(data_dir, "KS_coeffs.bin")

        evals = []
        rows = []
        cols = []
        vals = []
        with open(infile) as ifile:
            nspin, ntmb, nfvctr = [int(x) for x in next(ifile).split()[:3]]

            # Values
            for i in range(0, nfvctr):
                evals.append(float(next(ifile).split()[0]))

            # Vectors
            for line in ifile:
                coeff, j, i = line.split()[:3]
                rows.append(int(j)-1)
                cols.append(int(i)-1)
                vals.append(float(coeff))

        mat = coo_matrix((vals, (rows, cols)), shape=(ntmb, nfvctr))

        return csc_matrix(mat), evals

    def get_matrix_kxs(self, log):
        """
        Computes the matrix K*S, the mulliken version of the density matrix,
        and loads it into memory.

        log (Logfile): instance of a Logfile class

        Returns:
          (scipy.sparse.csc_matrix): the matrix K*S
        """
        from BigDFT.Spillage import serial_compute_puritybase
        from scipy.sparse import csc_matrix
        from scipy.io import mmread, mmwrite
        from os.path import join

        def generate_kxs(kxsfile, k_and_s):
            kfile, sfile = k_and_s
            # Then compute with python
            kxs = serial_compute_puritybase(sfile.obj, kfile.obj)
            mmwrite(kxsfile, kxs)

        kxsfile = join(_get_datadir(log), 'kxs_sparse.mtx')

        try:
            kxs = self.derived_quantity(
                      kxsfile, log, ['density_kernel', 'overlap'],
                      generate_kxs)
        except IOError:
            kxs = self.derived_quantity(
                      kxsfile, log,
                      ['density_kernel_sparse', 'overlap_sparse'],
                      generate_kxs)

        # validate the node and read the matrix
        kxs.validate()

        return csc_matrix(mmread(kxsfile))

    def derived_quantity(self, name, log, files, generator):
        """
        Defines a quantity that can be derived from the ccs files
        in the files list. Requires the generator function.
        """
        from futile.Utils import Node, more_recent_than_parent
        from os.path import join, basename

        def more_recent_than_logfile(filename):
            logfile = join(log.srcdir, basename(log.label))
            return more_recent_than_parent(filename, logfile)

        def validate_all(h_and_s):
            ok = True
            for m in h_and_s:
                ok = ok and m.validate()
                if not ok:
                    break
            return ok

        def more_recent_than_parents(spillfile, node):
            ok = True
            for m in node:
                ok = ok and more_recent_than_parent(spillfile, m.obj)
                if not ok:
                    break
            return ok

        newfile = join(_get_datadir(log), name)
        # in case there is no need to rerun the generation
        newnode = Node(newfile, valid_if=more_recent_than_logfile)
        if newnode.valid:
            return newnode

        nodefiles = [self.ccs_node(log, f) for f in files]

        node = Node(nodefiles, valid_if=validate_all)

        newnode = Node(newfile)
        newnode.requires(node, generator=generator)
        newnode.valid_if(more_recent_than_parents)

        try:
            node.validate()
        except AttributeError:
            raise IOError()

        return newnode

    def ccs_node(self, log, matrixname):
        """
        Defines the  protocol to validate the ccs matrix file
        """
        from os.path import join, isfile
        from futile.Utils import Node, non_null_size

        data_dir = _get_datadir(log)
        mtxfile = join(data_dir, matrixname+'.mtx')
        if isfile(mtxfile):  # matrix market output
            return Node(mtxfile, valid_if=non_null_size)
        if log.log["lin_general"]["output_mat"] == 4:
            mpifile = join(data_dir, matrixname+".mpi")
        else:
            mpifile = None
        txtfile = join(data_dir, matrixname+".txt")
        ccsfile = join(data_dir, matrixname+".ccs")
        return file_dependency_queue(self, ccsfile, txtfile, mpifile=mpifile)

    def get_matrix_sinvh(self, log):
        """
        Computes the matrix S^{-1}*H, the mulliken version of the spillage
        matrix, and loads it into memory.

        log (Logfile): instance of a Logfile class

        Returns:
          (scipy.sparse.csc_matrix): the matrix S^{-1}*H
        """
        from BigDFT.Spillage import serial_compute_spillagebase
        from scipy.sparse import csc_matrix
        from scipy.io import mmread, mmwrite
        from os.path import join

        def generate_sinvh(spillfile, h_and_s):
            houtfile, soutfile = h_and_s
            # Then compute with python
            sinvh = serial_compute_spillagebase(soutfile.obj, houtfile.obj)
            mmwrite(spillfile, sinvh)

        spillfile = join(_get_datadir(log), 'sinvh_sparse.mtx')

        try:
            sinvh = self.derived_quantity(spillfile, log,
                                          ['hamiltonian', 'overlap'],
                                          generate_sinvh)
        except IOError:
            sinvh = self.derived_quantity(spillfile, log,
                                          ['hamiltonian_sparse',
                                           'overlap_sparse'],
                                          generate_sinvh)

        # validate the node and read the matrix
        sinvh.validate()

        return csc_matrix(mmread(spillfile))

    def _get_matrix(self, name, log):
        from os.path import join, isfile
        from BigDFT.Spillage import read_ccs
        from scipy.io import mmread
        from scipy.sparse import csc_matrix
        # Define the input files.
        data_dir = _get_datadir(log)

        # Check for mtx file
        mtxfile = join(data_dir, name+".mtx")
        if isfile(mtxfile):
            return csc_matrix(mmread(mtxfile))

        soutfile = join(data_dir, name+".ccs")

        # Convert if necessary
        if not isfile(soutfile):
            sfile = join(data_dir, name+".txt")
            # Convert to text format if necessary
            if log.log["lin_general"]["output_mat"] == 4:
                infile = join(data_dir, sfile.replace(".txt", ".mpi"))
                if not isfile(sfile):
                    self.convert_matrix_format(conversion="binary_to_bigdft",
                                               infile=infile, outfile=sfile)
            # Convert to CCS
            soutfile = join(data_dir, name+".ccs")
            self.convert_matrix_format(conversion="bigdft_to_ccs",
                                       infile=sfile, outfile=soutfile)

        return read_ccs(soutfile)

    def get_matrix_s(self, log):
        """
        Read the overlap matrix into memory.

        log (Logfile): instance of a Logfile class

        Returns:
          (scipy.sparse.csc_matrix): the matrix S
        """
        return self._get_matrix("overlap_sparse", log)

    def get_matrix_h(self, log):
        """
        Read the hamiltonian matrix into memory.

        log (Logfile): instance of a Logfile class

        Returns:
          (scipy.sparse.csc_matrix): the matrix H
        """
        return self._get_matrix("hamiltonian_sparse", log)

    def get_matrix_k(self, log):
        """
        Read the density matrix into memory.

        log (Logfile): instance of a Logfile class

        Returns:
          (scipy.sparse.csc_matrix): the matrix K
        """
        return self._get_matrix("density_kernel_sparse", log)

    def _invoke_command(self, command, **kwargs):
        from futile.Utils import option_line_generator
        _system_command(command, option_line_generator(**kwargs))


def file_dependency_queue(btool, ccsfile, txtfile, mpifile=None):
    from futile.Utils import Node, non_null_size, more_recent_than_parent

    def binary_to_bigdft(node, parent):
        btool.convert_matrix_format(conversion="binary_to_bigdft",
                                    infile=parent, outfile=node)

    def bigdft_to_ccs(node, parent):
        btool.convert_matrix_format(conversion="bigdft_to_ccs",
                                    infile=parent, outfile=node)

    if mpifile is not None:
        mpi = Node(mpifile, valid_if=non_null_size)
        txt = Node(txtfile)
        txt.requires(mpi, generator=binary_to_bigdft)
        txt.valid_if(more_recent_than_parent)
    else:
        txt = Node(txtfile, valid_if=non_null_size)

    ccs = Node(ccsfile)
    ccs.requires(txt, generator=bigdft_to_ccs)
    ccs.valid_if(more_recent_than_parent)

    return ccs


def _compute_purity_values(kxs, charges, frag_indices):
    """
    Computes the actual purity values.

    Args:
      kxs (scipy.sparse.csc): K*S
      charges (dict): the charge to divide by for each framgent.
      frag_indices (dict): indices associated with each fragment.

    Returns:
      (dict): purity of each fragment.
    """
    from numpy import trace
    purity_values = {}
    for id in frag_indices:
        indices = frag_indices[id]
        smat = kxs[:, indices]
        smat = 0.5 * smat[indices, :]
        purity_values[id] = 2.0 * trace(
            (smat.dot(smat) - smat).todense()) / charges[id]

    return purity_values


def superunits_purities(bo, pv, q_units, superunits, q_superunits):
    """
    Compute the purity values of the superunits described as a unification of
    the units

    Args:
       bo (matrix-like): Fragment Bond Orders of the Units
       pv (array-like): Purity values of the Units
       superunits (dict): lookup dictionary containing the list
         of units per superunit
       q_units (array-like): charges of the units
       q_superunits (array-like): charges of the superunits

    Returns:
       dict: purities of the superunits
    """
    purities = {}
    for fragid, atlist in superunits.items():
        purity = 0

        # Sum the bond orders
        for a1 in atlist:
            for a2 in atlist:
                if a1 != a2:
                    purity += bo[a1][a2]

        # Sum the un-normalized atomic purity values
        for a in atlist:
            charge = q_units[a]  # sum([x.nel for x in atsys["ATOM:"+str(a)]])
            purity += pv[a] * charge / 2.0

        # Normalize
        purity *= 2.0 / q_superunits[fragid]
        # (sum([x.nel for x in resys[fragid]]))
        purities[fragid] = purity
    return purities


def units_quadratic_traces(unitlist1, unitlist2, frag_indices, mat1, mat2,
                           a1=1.0, a2=1.0):
    from numpy import einsum
    # Double loop over the pairs to compute.
    interaction = {}
    for fragid1 in unitlist1:
        interaction[fragid1] = {}
        idx1 = frag_indices[fragid1]
        lmat1 = mat1[:, idx1].todense()  # smat1
        rmat1 = mat2[idx1, :].todense()  # smat2.T

        for fragid2 in unitlist2:
            idx2 = frag_indices[fragid2]

            # Optimization for the case where the indices are consecutive
            # values.
            sidx2 = sorted(idx2)
            mindx = min(sidx2)
            maxdx = max(sidx2)
            if len(sidx2) == maxdx - mindx + 1:
                lmat2 = lmat1[mindx:maxdx+1, :]  # smat12
                rmat2 = rmat1[:, mindx:maxdx+1]  # smat21.T
            else:
                lmat2 = lmat1[idx2, :]  # smat12
                rmat2 = rmat1[:, idx2]  # smat21.T

            interaction[fragid1][fragid2] = a1*a2*einsum("ij,ji->", lmat2,
                                                         rmat2)
    return interaction


def superunits_quadratic_quantities(bo, superunits):
    """
    Compute quantities that transforms like the bond orders of the superunits
    described as a unification of the units

    Args:
       bo (matrix-like): Quantities like Fragment Bond Orders of the Units
       superunits (dict): lookup dictionary containing the list
         of units per superunit

    Returns:
       dict: quantities of the superunits
    """
    bo_new = {}
    for f1, atlist1 in superunits.items():
        for f2, atlist2 in superunits.items():
            order = 0
            for a1 in atlist1:
                for a2 in atlist2:
                    order += bo[a1][a2]
            bo_new.setdefault(f1, {})[f2] = order
    return bo_new


def systems_heatmap(data, restrict_to=None, axs=None, columns=None, **kwargs):
    """
    Create a heatmap for a set of systems.

    Args:
        data (dict): a dictionary mapping system names to another dictionary
          which maps fragment ids to property values.
        restrict_to (list): a list of lists saying what fragments we should
          draw. It is a list of lists instead of just a list because this
          way we can put a separator between values (for example, when making
          a jump in the sequence).
        axs (matplotlib.axis): the axis to draw on.
        columns (list): the order of the columns on the x axis
        kwargs (dict): any extra argument you wish to pass to matplotlib's
          imshow command.

    Returns:
        a reference to the imshow value.
    """
    from numpy import zeros
    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    # Handle optional axis
    if axs is None:
        fig, axs = plt.subplots(1, 1, figsize=(12, 8))

    # Handle the restrict_to option.
    if restrict_to is not None:
        keylist = [x for x in [y for y in restrict_to]]
        keylist = [x for y in restrict_to for x in y]
    else:
        key = list(data.keys())[0]
        keylist = list(data[key])

    cols = columns if columns is not None else data.keys()

    # Create the matrix of values and plot it.
    mat = zeros((len(keylist), len(cols)))
    for i, keyi in enumerate(cols):
        for j, keyj in enumerate(keylist):
            mat[j, i] = data[keyi][keyj]
    im = axs.imshow(mat, **kwargs)

    # Labels
    if restrict_to is not None:
        tickloc = []
        ticklab = []
        i = 0
        for x in restrict_to:
            for y in x:
                tickloc.append(i)
                ticklab.append(y)
                i += 1
            if x == restrict_to[-1]:
                break
            tickloc.append(i-0.5)
            ticklab.append("-------")
        axs.set_yticks(tickloc)
        axs.set_yticklabels(ticklab, family='monospace')
    else:
        axs.set_yticks(range(mat.shape[0]))
        axs.set_yticklabels(keylist, family='monospace')
    axs.set_xticks(range(mat.shape[1]))
    axs.set_xticklabels(list(cols))

    for tick in axs.get_xticklabels():
        tick.set_rotation(75)

    # Draw the colorbar
    divider = make_axes_locatable(axs)
    cax = divider.append_axes("right", size="2%", pad=0.15)
    plt.colorbar(im, cax=cax)

    # Draw lines that separate out the boxes.
    for i in range(mat.shape[0]-1):
        axs.axhline(i+0.45, color='k', linestyle='-', linewidth=0.5)
    for i in range(mat.shape[1]-1):
        axs.axvline(i+0.48, color='k', linestyle='-', linewidth=0.5)

    # Not having a fixed aspect is good for larger plots.
    axs.set_aspect("auto")

    return im


def _example():
    """
    Postprocessing Example
    """
    from BigDFT.Systems import System, FragmentView
    from BigDFT.Fragments import Fragment
    from BigDFT.IO import XYZReader
    from BigDFT.Calculators import SystemCalculator
    from BigDFT.Inputfiles import Inputfile
    from scipy.linalg import eigh
    from copy import deepcopy

    # Create a system
    sys = System()
    sys["FRA:0"] = Fragment()
    with XYZReader("CH2") as ifile:
        for at in ifile:
            sys["FRA:0"].append(at)
    sys["FRA:1"] = Fragment()
    with XYZReader("CH3") as ifile:
        for at in ifile:
            sys["FRA:1"].append(at)
    sys["FRA:1"].translate([0, 0, -3])
    sys["FRA:2"] = deepcopy(sys["FRA:0"])
    sys["FRA:2"].translate([0, 0, 3])
    sys["FRA:2"].rotate(y=150, units="degrees")
    sys["FRA:3"] = deepcopy(sys["FRA:0"])
    sys["FRA:3"].translate([3, 0, 1.5])

    print(list(sys))

    # Run a calculation. `write_support_function_matrices` and `linear` are
    # key. You also should be sure to set the atom multipoles.
    inp = Inputfile()
    inp.set_xc("PBE")
    inp.set_hgrid(0.4)
    inp.write_support_function_matrices()
    inp["import"] = "linear"

    code = SystemCalculator()
    code.update_global_options(verbose=False)
    log = code.run(input=inp, posinp=sys.get_posinp(), run_dir="work")
    sys.set_atom_multipoles(log)

    # Create the post processing tool.
    from BigDFT.PostProcessing import BigDFTool
    btool = BigDFTool()

    # Purity
    purity = btool.run_compute_purity(sys, log)
    print(purity)

    # Charges
    charges = {fragid: sum(at.nel for at in frag)
               for fragid, frag in sys.items()}

    # Bond Orders
    bo = btool.fragment_bond_order(sys, sys.keys(), sys.keys(), log)

    # These three things define a fragment view.
    view = FragmentView(purity, bo, charges)

    # Auto Fragmentation
    mapping = btool.auto_fragment(sys, view, 0.10)
    print(mapping)

    # This defines a new view.
    new_view = view.refragment(mapping)
    print(new_view.purities)

    # Eigenvalues.
    H = btool.get_matrix_h(log)
    S = btool.get_matrix_s(log)
    w = eigh(H.todense(), b=S.todense(), eigvals_only=True)
    print(w)


if __name__ == "__main__":
    _example()



"""
This module contains some wrappers for using PSI4 to perform calculations.

https://psicode.org/
"""
from BigDFT.Calculators import Runner
from futile.Utils import write as safe_print


def bigdft_to_psi4(sysA, sysB=None, spinA=0, multiplicityA=1, spinB=0,
                   multiplicityB=1):
    """
    Create a PSI4 dimer molecule which can be used for SAPT.

    If multiple molecules are specified, we create a Dimer system.

    Args:
        sysA (BigDFT.Systems.System): the first molecule.
        sysB (BigDFT.Systems.System): the second molecule (optional).
        spinA (int): the spin of the first molecule.
        spinB (int): the spin of the second molecule.
        multiplicityA (int): the multiplicity for unpaired electrons.
        multiplicityB (int): the multiplicity for unpaired electrons.

    Returns:
        (psi4.core.Molecule): the psi4 geometry.
    """
    from BigDFT.IO import write_xyz
    from psi4 import geometry
    # py2 workaround
    from sys import version_info
    if version_info[0] < 3:
        from io import BytesIO as StringIO
    else:
        try:
            from io import StringIO
        except ImportError:
            from StringIO import StringIO

    # Create the right string representations
    strA = StringIO()
    write_xyz(sysA, strA)
    xyzA = str(spinA) + " " + str(multiplicityA) + "\n"
    xyzA += "\n".join(strA.getvalue().split("\n")[2:])

    if sysB is not None:
        strB = StringIO()
        write_xyz(sysB, strB)
        xyzB = str(spinB) + " " + str(multiplicityB) + "\n"
        xyzB += "\n".join(strB.getvalue().split("\n")[2:])

    if sysB is None:
        return geometry("\n"+xyzA+"\nunits angstrom\n")
    else:
        return geometry("\n"+xyzA+"--\n"+xyzB+"units angstrom\n")


class PSI4Calculator(Runner):
    """
    Perform a calculation on a given system using the PSI4 code.

    Note that if you intend to use an SAPT method, you need to pass a
    system which is composed of exactly two fragments.

    PSI4 has a number of different actions, ab initio methods, and basis
    sets. Be sure to specify each of these to the run command.
    """
    import os

    def __init__(self, omp=os.environ.get('OMP_NUM_THREADS', '1'),
                 skip=False, verbose=True):
        # Use the initialization from the Runner class (so all options inside
        # __global_options)
        self.mol = None
        Runner.__init__(self, omp=str(omp), dry_run=False, skip=skip,
                        verbose=verbose)

    def pre_processing(self):
        """
        Process local run dictionary to create the input directory and identify
        the command to be passed

        Returns:
            :py:class:`dict`: dictionary containing the command to be passed to
            :meth:`process_run`
        """
        from psi4 import set_options, energy, optimize, frequency
        from psi4.core import set_output_file
        from BigDFT.Systems import System
        from os.path import join

        # Check Arguments
        try:
            self.sys = self.run_options["sys"]
        except KeyError:
            raise ValueError("sys= must be provided as an argument")

        try:
            action = self.run_options["action"]
        except KeyError:
            raise ValueError("You must specify a valid action=(energy,"
                             " optimize, frequency")

        try:
            method = self.run_options["method"]
        except KeyError:
            raise ValueError("You must specify an ab initio method=")

        try:
            basis = self.run_options["basis"]
        except KeyError:
            raise ValueError("You must specify a basis=")

        # Run directory
        self._ensure_run_directory()

        # Check skip
        if self.run_options["skip"]:
            try:
                log = PSI4Logfile(self.sys, self._get_logname(True),
                                  action, method)
                return {"command": None}
            except ValueError:  # invalid logfile, so we can't skip.
                pass
            except IOError:  # no logfile, so we can't skip.
                pass

        # Convert system
        if "sapt" in method:
            keys = list(self.sys)
            if len(keys) != 2:
                raise ValueError("For SAPT method, your system must be"
                                 " composed of exactly two fragments.")
            sysA = System()
            sysA[keys[0]] = self.sys[keys[0]]
            sysB = System()
            sysB[keys[1]] = self.sys[keys[1]]
            mol = bigdft_to_psi4(sysA=sysA, sysB=sysB)
        else:
            mol = bigdft_to_psi4(sysA=self.sys)

        # Set extra options.
        if "psi4_options" in self.run_options:
            set_options(self.run_options["psi4_options"])
        set_output_file(self._get_logname(True), False)

        # Instead of a command, we return a closure.
        if action == "energy":
            def cmd(method, mol):
                energy(method, molecule=mol)
        elif action == "optimize":
            def cmd(method, mol):
                optimize(method, molecule=mol)
        elif action == "frequency":
            def cmd(method, mol):
                frequency(method, molecule=mol)

        return {"command": cmd(join(method, basis), mol)}

    def process_run(self, command):
        """
        Run the psi4 command.
        """
        from os import environ, system
        # Set the number of omp threads only if the variable is not present
        # in the environment
        if 'OMP_NUM_THREADS' not in environ:
            environ['OMP_NUM_THREADS'] = self.run_options['omp']

        if self.run_options['verbose']:
            if self.run_dir != '.':
                safe_print('Run directory', self.run_dir)

        # Run the command
        if command is not None:
            command()

        return {'logname': self._get_logname(True)}

    def post_processing(self, logname, command):
        """
        Post processing the calculation.

        Returns:
            (BigDFT.Interop.PSI4Interop.PSI4Logfile): a representation of the
            logfile.
        """
        try:
            return PSI4Logfile(self.sys, fname=logname,
                               action=self.run_options["action"],
                               method=self.run_options["method"])
        except IOError:
            raise ValueError("The logfile (", logname, ") does not exist.")

    def _ensure_run_directory(self):
        from futile.Utils import ensure_dir
        run_dir = self.run_options.get('run_dir', '.')
        # Create the run_dir if not exist
        if ensure_dir(run_dir) and self.run_options['verbose']:
            safe_print("Create the sub-directory '%s'" % run_dir)

        self.run_dir = run_dir

    def _get_logname(self, full):
        from os.path import join
        name = self.run_options.get("name", "sapt") + ".log"
        if full:
            run_dir = self.run_options.get('run_dir', '.')
            return join(run_dir, name)
        else:
            return name


class PSI4Logfile(dict):
    """
    A logfile wrapper for an SAPT calculation.

    This logfile inherits from a dictionary, as the properties we get out
    of a calculation depend on the type of calculation performed.

    Args:
        fname (str): the name of the file to process.
        action (str): the action that was used when running SAPT (energy,
          optimize, frequency).
        method (str): the scf method used.
    """
    def __init__(self, sys, fname, action, method):
        dict.__init__({})
        self["energy"] = {}

        self.sys = sys

        # The logfile, broken up into a list of lines.
        self.log = self._log_to_str(fname)

        if "sapt" in method:
            self._extract_sapt_energies(method)
        elif "energy" in action:
            self._extract_total_energy()
        elif "optimize" in action:
            self._extract_optimization_energies()
            self._extract_optimization_geometries()

    def _log_to_str(self, fname):
        with open(fname) as ifile:
            lstr = "".join([x for x in ifile])
        return lstr.split("\n")

    def _look_for(self, goal):
        for offset, line in enumerate(self.log):
            if goal in line:
                break
        if offset > len(self.log):
            raise ValueError("Couldn't find property in logfile: ", goal)
        return offset

    def _extract_optimization_energies(self):
        self["max force"] = []
        self["energy"]["optimize"] = []
        offset = self._look_for("==> Optimization Summary <==")
        for line in self.log[offset+6:]:
            if "---" in line:
                break
            split = line.split()
            self["max force"].append(float(split[3]))
            self["energy"]["optimize"].append(float(split[1]))
        self["energy"]["total"] = self["energy"]["optimize"][-1]

    def _extract_optimization_geometries(self):
        from BigDFT.Systems import System
        from BigDFT.Fragments import RotoTranslation
        from copy import deepcopy

        # Read out the posinps
        posinps = []
        geomcount = 0
        for i in range(len(self.log)):
            line = self.log[i]
            if "==> Geometry <==" in line:
                # It is printed twice, once for scf and one for gradient.
                geomcount += 1
                if geomcount % 2 == 0:
                    continue
                # Extract
                geom = {"positions": [], "units": "angstroem"}
                j = 9
                while(True):
                    split = self.log[i + j].split()
                    if len(split) == 0:
                        break
                    atom = {split[0].title(): [float(x) for x in split[1:4]]}
                    geom["positions"].append(atom)
                    j += 1
                posinps.append(geom)

        # Convert the posinps into first class system objects.
        self["positions"] = []
        for pos in posinps:
            self["positions"].append(deepcopy(self.sys))
            self["positions"][-1].update_positions_from_dict(pos)

        # PSI4 does rototranslations of the geometry, so we need to undo it.
        rt = {}
        for fragid in self.sys:
            rt[fragid] = RotoTranslation(self["positions"][0][fragid],
                                         self.sys[fragid])
        for osys in self["positions"]:
            for fragid in self.sys:
                osys[fragid] = rt[fragid].dot(osys[fragid])

    def _extract_total_energy(self):
        for line in self.log:
            if "Total Energy =" in line:
                self["energy"]["total"] = float(line.split()[-1])
        if "total" not in self["energy"]:
            raise ValueError("Could you find total energy in log file.")

    def _extract_sapt_energies(self, method):
        def _check_vals():
            if any([x not in vals for x in ["total", "exchange", "induction",
                                            "dispersion", "electrostatics"]]):
                raise ValueError("Could not find sapt energies in log file")

        # Get main computation
        vals = {}
        offset = self._look_for("SAPT Results")
        for line in self.log[offset:]:
            if "Electrostatics" in line:
                vals["electrostatics"] = float(line.split()[1]) / 1000.0
            elif "Exchange" in line:
                vals["exchange"] = float(line.split()[1]) / 1000.0
            elif "Induction" in line:
                vals["induction"] = float(line.split()[1]) / 1000.0
            elif "Dispersion" in line:
                vals["dispersion"] = float(line.split()[1]) / 1000.0
            elif "Total" in line:
                vals["total"] = float(line.split()[2]) / 1000.0
                break

        _check_vals()
        self["energy"][method] = vals

        # Extract special recipe version
        vals = {}
        for line in self.log:
            if "Total sSAPT0" in line:
                vals["total"] = float(line.split()[2]) / 1000.0
            if "Exchange sSAPT0" in line:
                vals["exchange"] = float(line.split()[2]) / 1000.0
            if "Induction sSAPT0" in line:
                vals["induction"] = float(line.split()[2]) / 1000.0
            if "Dispersion sSAPT0" in line:
                vals["dispersion"] = float(line.split()[2]) / 1000.0
            if "Electrostatics sSAPT0" in line:
                vals["electrostatics"] = float(line.split()[2]) / 1000.0

        _check_vals()
        self["energy"]["sSAPT0"] = vals


def _example():
    """Example of using PSI4 interoperability"""
    from BigDFT.IO import XYZReader
    from BigDFT.Systems import System
    from BigDFT.Fragments import Fragment
    from os.path import join
    from copy import deepcopy

    # Create a system.
    reader = XYZReader(join("Database", "XYZs", "He.xyz"))
    fsys = System()
    fsys["FRA:1"] = Fragment(xyzfile=reader)
    fsys["FRA:2"] = deepcopy(fsys["FRA:1"])
    fsys["FRA:2"].translate([-2, 0, 0])

    # Create a calculator.
    code = PSI4Calculator()
    log = code.run(sys=fsys, action="energy", basis="jun-cc-pvdz",
                   psi4_options={"scf_type": "direct"}, method="scf",
                   name="test")
    print(log)

    # The raw logfile is also available.
    print(log.log[4])

    # Geometry optimization.
    log = code.run(sys=fsys, action="optimize", basis="jun-cc-pvdz",
                   method="scf", name="test-opt")
    print(log)

    # SAPT
    log = code.run(sys=fsys, action="energy", basis="jun-cc-pvdz",
                   method="sapt0", name="test-sapt")
    print(log)


if __name__ == "__main__":
    _example()



"""
This module contains some wrappers for using XTB to perform calculations.

https://xtb-docs.readthedocs.io/en/latest/contents.html
"""
from BigDFT.Calculators import Runner
from futile.Utils import write as safe_print


class XTBCalculator(Runner):
    """
    A calculator which can drive simulations using the XTB program.

    XTB uses a mixed approach for control. First, you can pass command line
    arguments to the program. Second, you can modify the input file. We
    mirror this split here. Command line arguments can be passed as arguments
    to the run command. To modify the input file, the run command accepts
    an argument `inp` which is a dictionary. This dictionary is written
    to the input file (in the appropriate format). See the example for
    more details.

    https://xtb-docs.readthedocs.io/en/latest/contents.html
    """
    import os

    def __init__(self, omp=os.environ.get('OMP_NUM_THREADS', '1'),
                 dry_run=False, skip=False, verbose=True):
        # Use the initialization from the Runner class (so all options inside
        # __global_options)
        Runner.__init__(self, omp=str(omp), dry_run=dry_run, skip=skip,
                        verbose=verbose)
        # Build the command setting the number of omp threads
        self.command = 'xtb'
        safe_print(
            'Initialize a Calculator with OMP_NUM_THREADS=%s and command %s' %
            (self._global_options['omp'], self.command))

    def pre_processing(self):
        """
        Process local run dictionary to create the input directory and identify
        the command to be passed

        Returns:
            :py:class:`dict`: dictionary containing the command to be passed to
            :meth:`process_run`
        """
        self._ensure_run_directory()
        sys = self.run_options["sys"]
        inp = self.run_options.get("inp", {})

        # Write out the geometry file
        with open(self._get_inputfilename(), "w") as ofile:
            if sys.cell.get_boundary_condition() != "free":
                if any([sys.cell[x, x] == float("inf") for x in range(3)]):
                    raise ValueError("Only 3D periodic or free boundary " +
                                     "condition is supported")
                ofile.write("$periodic 3\n")
                ofile.write("$lattice bohr \n")
                for i in range(3):
                    ofile.write("  ")
                    ofile.write(" ".join([str(sys.cell[i, j])
                                          for j in range(3)]))
                    ofile.write("\n")

            ofile.write("$coord\n")
            for fragid, frag in sys.items():
                for at in frag:
                    pos = at.get_position("bohr")
                    ofile.write("  ")
                    ofile.write(str(pos[0]) + " ")
                    ofile.write(str(pos[1]) + " ")
                    ofile.write(str(pos[2]) + " ")
                    ofile.write(at.sym.lower() + "\n")
            ofile.write("$end\n")

            constr = self.run_options.get("constraints", None)
            if constr:
                self._set_constraints(constr, ofile)

            for k1, v1 in inp.items():
                ofile.write("$"+k1+"\n")
                for k2, v2 in v1.items():
                    ofile.write("  "+k2+"="+str(v2)+"\n")
                ofile.write("$end\n")

        return {'command': self._get_command()}

    def process_run(self, command):
        """
        Run the xtb executable.
        """
        from os import environ, system
        # Set the number of omp threads only if the variable is not present
        # in the environment
        if 'OMP_NUM_THREADS' not in environ:
            environ['OMP_NUM_THREADS'] = self.run_options['omp']
        if 'MKL_NUM_THREADS' not in environ:
            environ['OMP_NUM_THREADS'] = self.run_options['omp']

        if self.run_options['verbose']:
            if self.run_dir != '.':
                safe_print('Run directory', self.run_dir)
            safe_print('Executing command: ', command)

        # Run the command
        system("cd " + self.run_dir + "; " + command)

        return {'logname': self._get_logname(True)}

    def post_processing(self, logname, command):
        """
        Post processing the calculation.

        Returns:
            (BigDFT.Interop.XTBInterop.XTBLogfile): a representation of the
            logfile.
        """
        try:
            return XTBLogfile(sys=self.run_options["sys"], logname=logname)
        except IOError:
            raise ValueError("The logfile (", logname, ") does not exist.")

    def _ensure_run_directory(self):
        from futile.Utils import ensure_dir
        run_dir = self.run_options.get('run_dir', '.')
        # Create the run_dir if not exist
        if ensure_dir(run_dir) and self.run_options['verbose']:
            safe_print("Create the sub-directory '%s'" % run_dir)

        self.run_dir = run_dir

    def _get_command(self):
        from copy import deepcopy
        from os.path import exists

        command = self.command
        command += " " + self.run_options.get('name', 'input') + ".tmol"

        # Check if it is a dry run
        if self.run_options['dry_run']:
            return command + " --define"

        if self._check_skip():
            return '''echo "skip"'''

        options = deepcopy(self.run_options)

        # Adjust the command line with options
        name = self.run_options.get("name", "xtbrun")
        command += " --namespace " + name

        kv_ops = ["opt", "gfnff", "md", "omd", "hess", "gfn"]
        kv_ops += ["alpb", "chrg", "gbsa", "metadyn", "path", "uhf"]

        for k in kv_ops:
            if k in options:
                if isinstance(options[k], bool) and options[k] is True:
                    command += " --"+str(k)
                elif not isinstance(options[k], bool):
                    command += " --"+str(k) + " " + str(options[k])
                options.pop(k)

        command += " > " + self._get_logname(False)

        return command

    def _check_skip(self):
        from os.path import exists, join
        if not self.run_options["skip"]:
            return False
        fpath = join(self.run_dir, self._get_logname(False))
        if not exists(fpath):
            return False
        with open(fpath) as ifile:
            for line in ifile:
                if "* finished run on" in line:
                    return True
        return False

    def _get_logname(self, full):
        from os.path import join
        name = self.run_options.get("name", "xtbrun") + ".log"
        if full:
            run_dir = self.run_options.get('run_dir', '.')
            return join(run_dir, name)
        else:
            return name

    def _get_inputfilename(self):
        from os.path import join
        name = self.run_options.get('name', '')
        run_dir = self.run_options.get('run_dir', '.')
        input_file = name + '.tmol' if name else 'input.tmol'
        return join(run_dir, input_file)

    def _set_constraints(self, constr, ofile):
        from itertools import groupby
        from operator import itemgetter
        clist = []
        i = 1
        for fragid, frag in constr.items():
            for at in frag:
                if at:
                    clist.append(i)
                i += 1

        clist.sort()

        if len(clist) > 0:
            ofile.write("$fix\n")
            ofile.write("  atoms: ")
            ranges = []
            # Create ranges from consecutive indices.
            for k, g in groupby(enumerate(clist), lambda x: x[0]-x[1]):
                trange = list(map(itemgetter(1), g))
                if len(trange) == 1:
                    ranges.append(str(trange[0]))
                elif len(trange) > 1:
                    ranges.append(str(trange[0]) + "-" + str(trange[-1]))
            ofile.write(",".join(ranges))
            ofile.write("\n$end\n")


class XTBLogfile():
    """
    This class stores the results of an XTB calculation which might be later
    post-processed.

    Attributes:
        sys (BigDFT.Systems.System): a handle to the original system used
            in the calculation.
        log (str): a string representation of the logfile from XTB.
        charges (list): charge values computed by XTB.
        opt_energy (list): a list of energy values from the geometry
            optimization procedure.
        opt_fnorm (list): a list of gradient norm values from the geometry
            optimization procedure.
        opt_traj (list): a list of BigDFT.Systems.System from the geometry
            optimization procedure.
        md_energy (list): a list of energy values from the MD simulation.
        md_fnorm (list): a list of gradient norm values from the MD simulation.
        md_traj (list): a list of BigDFT.Systems.System from the MD simulation.
    """
    def __init__(self, sys, logname, opt=False, md=False):
        self.sys = sys
        fpath = logname.replace(".log", "")
        with open(fpath+".log") as ifile:
            self.log = "".join(ifile)
        try:
            with open(fpath+".charges") as ifile:
                self.charges = [float(x) for x in ifile]
        except FileNotFoundError:
            self.charges = None
        self.opt_energy, self.opt_fnorm, self.opt_traj = \
            self._process_trajectory(fpath + ".xtbopt.log")
        self.md_energy, self.md_fnorm, self.md_traj = \
            self._process_trajectory(fpath + ".xtb.trj")
        self.optimized_geometry = \
            self._get_optimized_geometry(fpath + ".xtbopt.tmol")

    def _process_trajectory(self, fname):
        from BigDFT.Atoms import Atom
        try:
            with open(fname) as ifile:
                ene = []
                fnorm = []
                systems = []
                while(True):
                    try:
                        natoms = int(next(ifile))
                        eneline = next(ifile).split()
                        ene.append(float(eneline[1]))
                        fnorm.append(float(eneline[3]))

                        atlist = []
                        for i in range(natoms):
                            split = next(ifile).split()
                            pos = [float(x) for x in split[1:]]
                            sym = split[0].upper()
                            atlist.append(Atom(sym=sym, r=pos,
                                               units="angstroem"))
                        systems.append(self._match_atlist(atlist))
                    except StopIteration:
                        break
        except FileNotFoundError:
            return (None, None, None)

        return (ene, fnorm, systems)

    def _get_optimized_geometry(self, fname):
        from BigDFT.Atoms import Atom
        try:
            with open(fname) as ifile:
                atlist = []
                line = next(ifile)
                while("$coord" not in line):
                    line = next(ifile)
                line = next(ifile)
                while("$" not in line):
                    split = line.split()
                    pos = [float(x) for x in split[:3]]
                    sym = split[-1].upper()
                    atlist.append(Atom(sym=sym, r=pos,
                                       units="bohr"))
                    line = next(ifile)
                system = self._match_atlist(atlist)
        except FileNotFoundError:
            return None

        return system

    def _match_atlist(self, atlist):
        from BigDFT.Systems import System
        from BigDFT.Fragments import Fragment
        from copy import deepcopy

        optsys = System()
        i = 0
        for fragid, frag in self.sys.items():
            optsys[fragid] = Fragment()
            for at in frag:
                updated = deepcopy(at)
                updated.set_position(atlist[i].get_position())
                optsys[fragid] += Fragment([updated])
                i += 1

        return optsys

    @property
    def energy(self):
        """
        The total energy of the system.
        """
        for line in self.log.split("\n"):
            if "| TOTAL ENERGY " in line:
                return float(line.split()[3])
        else:
            raise ValueError("Total energy not available")

    @property
    def gradient_norm(self):
        """
        The total gradient of the system.
        """
        for line in self.log.split("\n"):
            if "| GRADIENT NORM " in line:
                return float(line.split()[3])
        else:
            raise ValueError("Gradient not available")

    @property
    def homo_lumo_gap(self):
        """
        The homo lumo gap of the system.
        """
        for line in self.log.split("\n"):
            if "| HOMO-LUMO GAP " in line:
                return float(line.split()[3])
        else:
            raise ValueError("Gap not available")

    @property
    def gfnff_qest(self):
        """
        When using the GFNFF option, no charges are computed. Instead, there
        are estimated charges when the topology is generated. This will
        extract those values.
        """
        qest = []
        for i, line in enumerate(self.log.split("\n")):
            if "metchar sp-hybrid imet pi  qest" in line:
                break
        else:
            raise ValueError("qest not found in log file")

        for line in self.log.split("\n")[i+1:]:
            split = line.split()
            if len(split) == 0:
                break
            qest += [float(line.split()[8])]

        return qest


def xtb_equilibrate(sys, calc, prod_time=20, **kwargs):
    """
    This function will automate the process of taking a system from zero
    temperature up to room temperature (298.15 kelvin).

    This routine will work by performing 10ps simulations at temperatures
    from 0 to 298.15 kelvin, with each simulation spaced by 10 degrees
    (310 ps total). Finally, the production run can be performed.

    Args:
        sys (BigDFT.Systems.System): the system to calculate.
        calc (BigDFT.Interop.XTBInterop.XTBCalculator): a calculator to use.
        prod_time (int): the number of production picoseconds to run at
            room temperature.
        **kwargs: any argument you would normally want to pass to the
            run command.
    Returns:
        (list, list): a list of positions along the trajectory and a list of
        energy values.
    """
    from copy import deepcopy
    positions = []
    energies = []
    param = kwargs.pop("inp", {})
    if "md" not in param:
        param["md"] = {}

    # Loop over temperature values
    usys = deepcopy(sys)
    for temp in list(range(0, 300, 10)) + [298.15]:
        param["md"]["temp"] = temp
        param["md"]["time"] = 10
        if temp == 0:
            param["md"]["restart"] = "false"
            log = calc.run(sys=usys, omd=True, inp=param, **kwargs)
        else:
            param["md"]["restart"] = "true"
            log = calc.run(sys=usys, md=True, inp=param, **kwargs)

        # Extract results
        positions += log.md_traj
        energies += log.md_energy
        usys = positions[-1]

    # Production Run
    param["md"]["time"] = prod_time
    param["md"]["restart"] = "true"
    param["md"]["temp"] = 298.15
    log = calc.run(sys=usys, md=True, inp=param, **kwargs)

    positions += log.md_traj
    energies += log.md_energy
    usys = positions[-1]

    return positions, energies


def _example():
    """Example of using XTB interoperability"""
    from BigDFT.IO import XYZReader
    from BigDFT.Systems import System
    from BigDFT.Fragments import Fragment
    from os.path import join
    from copy import deepcopy

    # Create a system.
    reader = XYZReader("Ar")
    fsys = System()
    fsys["FRA:1"] = Fragment(xyzfile=reader)
    fsys["FRA:2"] = deepcopy(fsys["FRA:1"])
    fsys["FRA:2"].translate([-2, 0, 0])

    # Create a calculator.
    code = XTBCalculator()
    log = code.run(sys=fsys, name="test", run_dir="work")

    # Print some values.
    print(log.energy)
    print(log.gradient_norm)
    print(log.homo_lumo_gap)
    print(log.charges)

    # Geometry optimization.
    log = code.run(sys=fsys, name="opt", opt=True, run_dir="work")
    sys2 = log.optimized_geometry
    # this also works
    sys2 = log.opt_traj[-1]
    print(log.opt_energy)
    print(log.opt_fnorm)

    # Recompute.
    log = code.run(sys=sys2, name="opt2", run_dir="work")
    print(log.energy)
    print(log.gradient_norm)
    print(log.homo_lumo_gap)

    # Molecular dynamics.
    inp = {}
    inp["md"] = {"temp": 300, "time": 2}
    logmd = code.run(sys=fsys, name="md", omd=True, inp=inp, run_dir="work")
    print(logmd.md_energy)
    print(logmd.md_fnorm)
    sys3 = logmd.md_traj[-1]


if __name__ == "__main__":
    _example()



"""
A module which contains the routines needed for computing the spillage values.
"""

#: Conversion between Atomic Units and Bohr
AU_to_A = 0.52917721092


class MatrixMetadata:
    """
    This class contains the information stored in the sparse matrix
    metadata file.

    Args:
      filename (str): the name of the metadata file to initialize from.

    Attributes:
      atoms (list): a list of atoms with their positions and associated
        basis functions.
      matdim (int): the dimension of the matrix.
    """
    def __init__(self, filename):
        from BigDFT.Atoms import Atom
        self.atoms = []
        symbol_lookup = []
        with open(filename, "r") as ifile:
            # Read the first line
            matinfo = next(ifile).split()
            self.matdim, natoms, ntypes = [int(x) for x in matinfo[:3]]
            # Units
            next(ifile)
            # skip geocode
            line = next(ifile)
            # skip shift?
            line = next(ifile)
            # get the symbol lookup information
            for i in range(0, ntypes):
                line = next(ifile).split()
                symbol_lookup.append(line[2])
            # Get the atom positions
            for i in range(0, natoms):
                line = next(ifile).split()
                adict = {}
                adict["sym"] = symbol_lookup[int(line[0]) - 1]
                adict["r"] = [float(x) for x in line[1:4]]
                # THIS IS BECAUSE THE METADATA FILE CURRENTLY PRINTS THE
                # WRONG UNITS!
                adict["units"] = "bohr"
                adict["indices"] = []
                self.atoms.append(Atom(adict))
            # Get the indices
            for i in range(0, self.matdim):
                line = next(ifile).split()
                self.atoms[int(line[0]) - 1]["indices"].append(i)

    def get_frag_indices(self, system):
        """
        Retrive the indices of the matrix associated with each fragment
        of a system.

        Args:
          system (Systems.System): the set of fragments to get the indices
            of.

        Return:
          (dict): a mapping from fragment to indices.
        """
        matching = system.compute_matching(self.atoms)

        frag_indices = {}
        for fragid in system:
            frag_indices[fragid] = []
            for at in matching[fragid]:
                frag_indices[fragid] += self.atoms[at]["indices"]

        return frag_indices


def _get_csc(filename):
    from scipy.sparse import csc_matrix
    from scipy.io import mmread

    if '.mtx' in filename:
        return csc_matrix(mmread(filename))
    else:
        return read_ccs(filename)


def serial_compute_puritybase(sfile, dfile):
    """
    This routine computes the matrix K*S using python.

    You will need this for computing the purity values.

    Args:
     sfile (str): the file name of the overlap matrix.
         Must be in ccs of mtx format.
     dfile (str): the file name of the density kernel.
         Must be in ccs of mtx format.

    Returns:
      (scipy.sparse.csc, scipy.sparse.csc): K*S
    """
    # Read from file
    smat = _get_csc(sfile)
    dmat = _get_csc(dfile)

    return dmat.dot(smat)


def serial_compute_spillagebase(sfile, hfile):
    """
    This routine computes the matrix S^{-1}*H using python.

    You will need this for computing the spillage values.

    Args:
     sfile (str): the file name of the overlap matrix.
         Must be in ccs of mtx format.
     hfile (str): the file name of the hamiltonian.
         Must be in ccs of mtx format.

    Returns:
      (scipy.sparse.csc, scipy.sparse.csc): Sinverse * H
    """
    # Read from file
    smat = _get_csc(sfile)
    hmat = _get_csc(hfile)

    # Invert
    sinvmat = invert_ccs(smat)

    return sinvmat.dot(hmat)


def invert_ccs(mat):
    """
    Invert a CSC matrix and restore a sparsity pattern such as it can be
    written without the need for large files.

    Args:
       mat (csc_matrix): the sparese matrix to invert

    Returns:
       csc_matrix: the inverse
    """
    from numpy import array, abs
    from scipy.sparse.linalg import inv
    # compute the inverse
    minv = inv(mat)
    # create the mask
    mask = array(abs(minv[minv.nonzero()]) < 1e-8)[0]
    r = minv.nonzero()[0][mask]
    c = minv.nonzero()[1][mask]
    # Filter the matrix.
    minv[r, c] = 0
    minv = 0.5 * (minv + minv.T)
    minv.eliminate_zeros()
    return minv


def read_ccs(fname):
    """
    Read a CCS matrix from file into a scipy csc matrix.

    This routine uses the CCS matrix files that CheSS generates.
    Args:
      fname (str): name of the file.

    Result:
      scipy.sparse.csc_matrix: the matrix in the file.
    """
    from scipy.sparse import csc_matrix

    data = []
    indices = []
    indptr = []

    with open(fname, "r") as ifile:
        # Read the meta data.
        line = next(ifile)
        split = line.split()
        matdim = int(split[0])
        nel = int(split[2])
        split = next(ifile).split()

        # Read in the index pointers.
        indptr = []
        indntx = [int(x) - 1 for x in split]
        indptr += indntx
        while len(indntx) == 10000:
            split = next(ifile).split()
            indntx = [int(x) - 1 for x in split]
            indptr += indntx

        # Read the indices.
        added = 0
        while(added < nel):
            split = next(ifile).split()
            values = [int(x) - 1 for x in split]
            indices.extend(values)
            added += len(values)

        # Read the data
        for line in ifile:
            data.append(float(line))

    matrix = csc_matrix((data, indices, indptr), shape=(matdim, matdim))
    return matrix



"""A module to describe information coming from ensemble averaging

"""


class Population():
    max_moments = 2

    @classmethod
    def load(cls, filename):
        from numpy import load
        try:
            pop_dict = load(filename)
        except ValueError:
            pop_dict = load(filename, allow_pickle=True, encoding="latin1")
        items = pop_dict.item()
        pop = cls(labels=items['feature_labels'])
        for d, w, l in zip(items['datas'], items['weights'],
                           items['sample_labels']):
            pop.append(d, weight=w, label=l)
        return pop

    def __init__(self, labels=None):
        self.cumsums = None
        self.datas = []
        self.sample_labels = []
        self.feature_labels = labels
        self.weights = []

    def append(self, data, weight=1.0, label=None):
        """
        Insert new data to the population.

        Args:
            data (array-like, pandas.DataFrame, dict): the data to append.
                In case of a dictionary, the data is internally converted into
                a dataframe.
            weight (float): the weight that the sample has in the population
            label (str): the label of the sample
        """
        from pandas import DataFrame
        if isinstance(data, dict):
            data_ = DataFrame(data)
            if self.feature_labels is None:
                self.feature_labels = data_.columns
        else:
            data_ = data
            if self.feature_labels is None:
                self.feature_labels = [str(i) for i in range(len(data))]

        sums = safe_moments(data, self.max_moments, weight)
        # [weight*data_, weight*data_**2, weight]
        self.weights.append(weight)
        if self.cumsums is None:
            self.cumsums = sums
        else:
            for isum in range(self.max_moments+1):
                self.cumsums[isum] = safe_add(self.cumsums[isum],
                                              safe_multiply_and_pow(sums[isum],
                                                                    1, 1))
                # self.cumsums[isum] += sums[isum]
        self.datas.append(data_)
        self.sample_labels.append(
            str(len(self.datas)-1) if label is None else label)

    def to_dataframe(self, sample_labels=None, feature_labels=None):
        """
        Convert the population into a pandas dataframe

        Args:
            sample_labels (list): overrides the sample labels
            feature_labels (list): overrides the feature labels

        Returns:
            pandas.Dataframe: the dataframe of the population
        """
        import pandas as pd
        if sample_labels is None:
            sample_labels = self.sample_labels
        if feature_labels is None:
            feature_labels = self.feature_labels
        dd = {sample: data for sample, data in zip(sample_labels, self.datas)}
        return pd.DataFrame.from_dict(dd, orient='index',
                                      columns=feature_labels)

    def to_dict(self):
        """
        Convert the population to a dictionary
        """
        return {att: getattr(self, att)
                for att in ['datas', 'feature_labels', 'sample_labels',
                            'weights']}

    def to_file(self, filename):
        """
        Dump the populations to a numpy file.

        Args:
            filename (str): the file for the dumping
        """
        from numpy import save
        pop_dict = self.to_dict()
        save(filename, pop_dict)

    def to_excel(self, filename=None, writer=None, sheets=['mean', 'std'],
                 prefix='', mappable=None):
        """
        Dump the population to an excel file.

        Args:
            filename (str): Name of the file to write the data to
            writer (pandas.ExcelWriter): the instance of the writer class.
                Useful to append more sheets to the same excel file
            sheets (list, str): list of the mehtods that will be written per
                each of the sheet of the file. It can be also a string, like
                "all" to write all the data of the population in a separate
                file
            prefix (str): prefix of the sheets to be employed
            mappable (func): a function to be applied to the data before
                writing them to the file
        Returns:
            pandas.ExcelWriter: the instance of the writer needed to create the
                file.
        """
        import pandas as pd
        if writer is None:
            writer = pd.ExcelWriter(filename, engine='xlsxwriter')
        if sheets == 'all':
            datas = self.datas
            shn = [prefix + '-' + str(lb) for lb in self.sample_labels]
        else:
            datas = [getattr(self, sht) for sht in sheets]
            shn = [prefix + '-' + sht for sht in sheets]
        for d, n in zip(datas, shn):
            df = d if mappable is None else mappable(d)
            df.to_excel(writer, sheet_name=n)
        return writer

    @property
    def full_weight(self):
        return self.cumsums[self.max_moments]

    @property
    def mean(self):
        return safe_multiply_and_pow(self.cumsums[0],
                                     1.0/self.full_weight, 1)

    @property
    def std(self):
        from numpy import sqrt, abs  # to avoid rundoff problems
        if len(self.datas) > 1:
            tmp1 = safe_sub(safe_multiply_and_pow(self.cumsums[1],
                                                  1.0/self.full_weight, 1),
                            safe_multiply_and_pow(self.mean, 1.0, 2))
            tmp1 = safe_unary_op(tmp1, abs)
            tmp1 = safe_unary_op(tmp1, sqrt)
            return tmp1
            # sqrt(abs(self.cumsums[1]/self.full_weight-self.mean**2))
        else:
            return 0.0


def remove_nan(df, transpose=False):
    df1 = df.dropna(how='all')
    dft = df1.transpose()
    df1 = dft.dropna(how='all')
    if transpose:
        return df1
    else:
        return df1.transpose()


def symmetrize_df(df1):
    """
    From a dataframe that should be asymmetrix matrix,
    construct the symmetrized dataframe
    """
    import numpy as np
    from pandas import DataFrame
    A = df1.values
    W = np.tril(A) + np.triu(A.T, 1)
    return DataFrame(W, columns=df1.columns, index=df1.index)


def reorder(dft, transpose=False):
    from BigDFT.IO import reorder_fragments
    dft1 = dft.reindex(index=reorder_fragments(dft.index))
    df1 = dft1.transpose()
    dft1 = df1.reindex(index=reorder_fragments(df1.index))
    if transpose:
        return dft1
    else:
        return dft1.transpose()


def clean_dataframe(df, symmetrize=True):
    """
    Symmetrize a dataframe and remove the NaN rows and columns

    Args:
       df (Dataframe)
       symmetrize (bool): symmetrize the dataframe if applicable
    Returns:
       Dataframe: the cleaned dataframe
    """
    dft = remove_nan(df, transpose=True)
    if symmetrize:
        dft = symmetrize_df(dft)
    return reorder(dft, transpose=True)


def stacked_dataframe(pop):
    """
    Construct a stacked dataframe with all the data of the population

    Warning:
        Weights are ignored, therefore the average value of such stacked
        dataframe may be different from the population mean.
    """
    from pandas import DataFrame as DF
    df = DF()
    for dd, name in zip(pop.datas, pop.sample_labels):
        df[name] = dd.stack()
    return df.T


def _convert_into_percent(weights):
    ntot = sum(weights)
    pcs = [(float(w)/ntot)*100.0 for w in weights]
    assert abs(sum(pcs) - 100.0) < 1.e-3
    return pcs


def weighted_dataframe(dfs, wgts):
    """
    Construct a single dataframe that include the provided weigths.
    Useful for all the situations where one wants to have a single view
    of a population which is weighted

    Args:
        dfs (list): list of dataframes to be weighted
        wgts(list):  weights to be included, should be of same length of dfs
    """
    from pandas import DataFrame
    from numpy import rint
    pcs = _convert_into_percent(wgts)
    df_tot = DataFrame()
    for df, pc in zip(dfs, pcs):
        for p in range(int(rint(pc))):
            df_tot = df_tot.append(df)
    return df_tot


def transverse_dataframe(population, target_features, target_samples):
    """
    Contruct a transverse dataframe that gathers the data from some target
    features and samples of a population. This may be useful to show
    the variability of some data for particular entries
    """
    df = stacked_dataframe(population)
    lookup = [(x, res) for x in target_samples for res in target_features]
    return df[lookup]


def concatenate_populations(populations, extra_labels={}):
    """
    Write a file containing the set of the populations we want to serialized

    Args:
         populations (dict): dictionary of the populations, labeled by the key
         extra_labels (dict): dictionary of the feature_labels
             and sample_labels of the populations
    Returns:
         pandas.Dataframe: the concatenated populations
    """
    from pandas import concat
    dfs = []
    for label, pop in populations.items():
        extra_lb = extra_labels.get(label, {})
        df = pop.to_dataframe(**extra_lb)
        df['label'] = label
        dfs.append(df)
    return concat(dfs)


def dump_populations(populations):
    """
    Dump a dictionary of populations in a archive

    Args:
         populations(dict): the dictionary of the populations

    Returns:
         set: the set of filenames (extension '.npy')
             which have been produced
    """
    files = []
    for label, pop in populations.items():
        filename = str(label) + '.npy'
        pop.to_file(filename)
        files.append(filename)
    return files


def safe_moments(data, order, weight):
    """
    Calculate a list of the moments of the distribution
    """
    moments = []
    for pw in range(order):
        moments.append(safe_multiply_and_pow(data, weight, pw+1))
    moments.append(weight)
    return moments


def safe_add(a, b):
    from pandas import DataFrame as DF
    try:
        return a - safe_multiply_and_pow(b, -1, 1)
    except Exception:
        return DF(a) + DF(b)


def safe_sub(a, b):
    return safe_add(a, safe_multiply_and_pow(b, -1, 1))


def safe_unary_op(data, op):
    """
    Apply the operation to the data in a dataframe-compatible way
    """
    from pandas import DataFrame as DF

    def _safe_op(x):
        try:
            return op(x)
        except Exception:
            return x

    try:
        return op(data)
    except Exception:
        try:
            return data.apply(_safe_op)
        except Exception:
            df = DF(data)
            return df.apply(op)


def safe_multiply_and_pow(data, a, pw):
    """
    Perform a multiplication and power expansion that is
    resilient to dataframes that contain sequences that are not
    floating-point number compatible
    """
    return safe_unary_op(data, _transform_pw(float(a), pw))


def _atimesx(a, x):
    if a == 1.0:
        return x
    else:
        return a*x


def _axpw(a, pw, x):
    if pw == 1:
        return _atimesx(a, x)
    else:
        return _atimesx(a, x**pw)


def _transform_pw(a=1.0, pw=1):
    from functools import partial
    return partial(_axpw, a, pw)


class ClusterGrammer():
    """
    A class that facilitates the use of the clustergrammer objects
    Args:
        df (pandas.DataFrame): the dataframe to represent the clustergrammer
    """
    def __init__(self, df):
        import clustergrammer_widget as cw
        self.df = df
        self.net = cw.Network(cw.clustergrammer_widget)
        self.net.load_df(df)

    def categorize(self, axis, cats):
        """
        Define categories to be displayed next to the axis elements

        Args:
            axis (str): 'row' or 'col'
            cats (dict): label: cats dictionary where cats is a dictionary
                where each key contains a list of fragments
        """
        recats = [{'title': tl, 'cats': c} for tl, c in cats.items()]
        self.net.add_cats(axis, recats)

    def represent_only(self, axis, elements):
        """
        Represent only the elements that are indicated on the given axis

        Args:
            axis (str): 'row' of 'col'
            elements (list): list of the elements to be represented on the
                given axis
        """
        from tempfile import NamedTemporaryFile as tmp
        # from os import remove
        import sys
        # redirect stdout
        old_stdout = sys.stdout
        ftmp = tmp()
        sys.stdout = ftmp
        self.net.filter_names(axis, elements)
        sys.stdout = old_stdout
        # remove(ftmp.name)

    def show(self):
        """
        Display the ClusterGrammer
        """
        self.net.cluster(enrichrgram=True)
        return self.net.widget()

    def publish(self, name):
        """
        Produce a link of the clustergrammer object that is
        in the public domain

        Args:
            name (str): name of the file to be created remotely
        """
        import requests
        filename = name+'.tsv'
        self.net.write_matrix_to_tsv(filename=filename)
        upload_url = 'http://amp.pharm.mssm.edu/clustergrammer/matrix_upload/'
        self.req = requests.post(upload_url,
                                 files={'file': open(filename, 'rb')})

    def publish_link(self):
        """
        Returns the URL of the published ClusterGrammer

        Returns:
            str: URL of the link
        """
        return self.req.text


class CData():
    """
    A class that facilitates the calls to data clustering algorithms.

    Args:
        datas (array-like): the data samples to perform the clustering to
    """
    def __init__(self, datas):
        self.datas = datas

    @property
    def lookup_nonnan(self):
        from numpy import isnan, where
        return where([not isnan(d) for d in self.datas[0]])

    @property
    def scaled_data(self):
        if not hasattr(self, '_scaled_data'):
            from sklearn.preprocessing import MinMaxScaler
            mms = MinMaxScaler()
            arr = self.datas.copy()
            mms.fit(arr)
            self._scaled_data = mms.transform(arr)
        return self._scaled_data

    def cluster_data(self, n_clusters):
        from sklearn.cluster import KMeans
        from sklearn.metrics import silhouette_samples, silhouette_score
        km = KMeans(n_clusters=n_clusters)
        km = km.fit(self.scaled_data)
        self.n_clusters = n_clusters
        self.cluster_labels = km.fit_predict(self.scaled_data)
        # Labeling the clusters
        self.cluster_centers = km.cluster_centers_
        # total distances of the clusters
        self.squared_distances = km.inertia_
        if self.n_clusters >= 2:
            # silhouette method (valid if nclusters >= 2)
            self.silhouette_avg = silhouette_score(self.scaled_data,
                                                   self.cluster_labels)
            # Compute the silhouette scores for each sample
            self.silhouette_values = silhouette_samples(self.scaled_data,
                                                        self.cluster_labels)
        return km

    def silhouette_plot(self, ax1=None):
        import numpy as np
        import matplotlib.cm as cm
        import matplotlib.pyplot as plt
        if ax1 is None:
            ax1 = plt.axes()
        # The 1st subplot is the silhouette plot
        # The silhouette coefficient can range from -1, 1
        ax1.set_xlim([-1, 1])
        # The (n_clusters+1)*10 is for inserting blank space between silhouette
        # plots of individual clusters, to demarcate them clearly.
        ax1.set_ylim([0, len(self.scaled_data) + (self.n_clusters + 1) * 10])
        y_lower = 10
        for i in range(self.n_clusters):
            # Aggregate the silhouette scores for samples belonging to
            # cluster i, and sort them
            ith_cluster_silhouette_values = \
                self.silhouette_values[self.cluster_labels == i]

            ith_cluster_silhouette_values.sort()

            size_cluster_i = ith_cluster_silhouette_values.shape[0]
            y_upper = y_lower + size_cluster_i

            color = cm.nipy_spectral(float(i) / self.n_clusters)
            ax1.fill_betweenx(np.arange(y_lower, y_upper),
                              0, ith_cluster_silhouette_values,
                              facecolor=color, edgecolor=color, alpha=0.7)

            # Label the silhouette plots with cluster numbers at the middle
            ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))

            # Compute the new y_lower for next plot
            y_lower = y_upper + 10  # 10 for the 0 samples

        ax1.set_title("The silhouette plot for the various clusters.")
        ax1.set_xlabel("The silhouette coefficient values")
        ax1.set_ylabel("Cluster label")

        # The vertical line for average silhouette score of all the values
        ax1.axvline(x=self.silhouette_avg, color="red", linestyle="--")

        ax1.set_yticks([])  # Clear the yaxis labels / ticks
        ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])
        return ax1

    # 2nd Plot showing the actual clusters formed
    def plot_clusters(self, feature_one=None, feature_two=None, axis=None):
        import matplotlib.cm as cm
        import matplotlib.pyplot as plt
        if axis is None:
            axis = plt.axes()
        one = list(self.datas.columns).index(feature_one)
        two = list(self.datas.columns).index(feature_one)
        name_one = str(feature_one)
        name_two = str(feature_two)
        # name_one, one, name_two, two = self.largest_features_ids(feature_one,
        #                                                          feature_two)
        colors = cm.nipy_spectral(
                    self.cluster_labels.astype(float) / self.n_clusters)
        axis.scatter(self.scaled_data[:, one], self._scaled_data[:, two],
                     marker='.', s=30, lw=0, alpha=0.7, c=colors,
                     edgecolor='k')
        # Draw white circles at cluster centers
        axis.scatter(self.cluster_centers[:, one],
                     self.cluster_centers[:, two], marker='o', c="white",
                     alpha=1, s=200, edgecolor='k')
        for i, c in enumerate(self.cluster_centers):
            axis.scatter(c[one], c[two], marker='$%d$' % i, alpha=1, s=50,
                         edgecolor='k')
        axis.set_title("The visualization of the clustered data.")
        axis.set_xlabel("Feature space for "+name_one)
        axis.set_ylabel("Feature space for "+name_two)
        return axis

    def plot_cluster_info(self, feature_one=None, feature_two=None):
        import matplotlib.pyplot as plt
        # Create a subplot with 1 row and 2 columns
        fig, (ax1, ax2) = plt.subplots(1, 2)
        fig.set_size_inches(18, 7)
        if self.n_clusters >= 2:
            self.silhouette_plot(ax1)
        self.plot_clusters(feature_one, feature_two, ax2)
        return plt

    def largest_features_ids(self, feature_one, feature_two):
        # name_one, name_two = self.get_largest_variances_names()
        if feature_one is not None:
            name_one = feature_one
        if feature_two is not None:
            name_two = feature_two
        labels = [self.feature_labels[i] for i, t in enumerate(
                                                   self.lookup_nonnan[0]) if t]
        one = labels.index(name_one)
        two = labels.index(name_two)
        return [name_one, one, name_two, two]

    def get_largest_variances_names(self):
        import numpy as np
        ordered_variances = np.argsort(self.std).tolist()
        ordered_variances.reverse()
        name_one = None
        name_two = None
        for iname in ordered_variances:
            if np.isnan(self.mean[iname]):
                continue
            if name_one is None:
                name_one = self.feature_labels[iname]
                continue
            if name_two is None:
                name_two = self.feature_labels[iname]
                break
        return name_one, name_two



"""
This module contains data structures for describing fragments. Fragments are
orders lists of atoms.
"""

from futile.Utils import write as safe_print
try:
    from collections.abc import MutableSequence
except ImportError:
    from collections import MutableSequence


def distance(i, j, cell=None):
    """
    Distance between fragments, defined as distance between center of mass

    Args:
      i (Fragment): first fragment.
      j (Fragment): second fragment.
      cell (array): an array describing the unit cell.

    Returns:
      (float): the distance between centers of mass.
    """
    import numpy
    vec = i.centroid - j.centroid
    if cell:
        per = numpy.where(cell > 0.0)
        for i, p in enumerate(per):
            vec -= cell[i] * int(round(vec[i] / cell[i]))
    return numpy.sqrt(numpy.dot(vec, vec.T))


def pairwise_distance(i, j, cell=None):
    """
    Distance between fragments, as defined by the distance between their two
    nearest atoms.

    Args:
      i (Fragment): first fragment.
      j (Fragment): second fragment.
      cell (array): an array describing the unit cell.

    Returns:
      (float): the pairwise distance between fragments.
    """
    if not cell:
        from scipy.spatial.distance import cdist
        pos_i = [at.get_position() for at in i]
        pos_j = [at.get_position() for at in j]

        dmat = cdist(pos_i, pos_j)

        dist = dmat.min()

    else:
        dist = distance(Fragment([i[0]]), Fragment([j[0]]), cell)
        for at_i in i:
            for at_j in j:
                new_d = distance(Fragment([at_i]), Fragment([at_j]), cell)
                if new_d < dist:
                    dist = new_d

    return dist


def lineup_fragment(frag):
    """
    Align the principal axis of inertia of the fragments along the
    coordinate axis. Also shift the fragment such as its centroid is zero.

    Args:
      (BigDFT.Fragments.Fragment): the fragment to transform.

    Returns:
      (BigDFT.Fragments.Fragment): the transformed fragment.
    """
    from numpy.linalg import eig

    # Shift the centroid to the origin.
    Shift = Translation(frag.centroid)
    Shift.invert()
    new_frag = Shift.dot(frag)

    # Align the principal axis of inertia are on the coordinate axis
    Imat = new_frag.ellipsoid()
    w, v = eig(Imat)
    Redress = Rotation(v.T)
    return Redress.dot(new_frag)


class RotoTranslation():
    """
    Define a transformation which can be applied to a fragment. This
    rotation is defined by giving this class two fragments, and the
    rototranslation between those fragments is automatically computed.

    Args:
      frag1 (BigDFT.Fragments.Fragment): the first position.
      frag2 (BigDFT.Fragments.Fragment): the second position.
    """
    def __init__(self, frag1, frag2):
        try:
            from BigDFT import wahba
            from numpy import matrix

            # Setup each fragment as a matrix of positions
            pos1 = matrix([at.get_position() for at in frag1])
            pos2 = matrix([at.get_position() for at in frag2])

            # Compute transformation
            self.R, self.t, self.J = wahba.rigid_transform_3D(pos1, pos2)
        except Exception as e:
            safe_print('Error', e)
            self.R, self.t, self.J = (None, None, 1.0e10)

    def dot(self, frag):
        """
        Apply the rototranslations on a fragment.

        Args:
          frag (BigDFT.Fragments.Fragment): the fragment to rototranslate.

        Return:
          (BigDFT.Fragments.Fragment): the rototranslated fragment.
        """
        from BigDFT import wahba as w
        from numpy import matrix, array
        from copy import deepcopy

        # Convert to a position matrix
        pos = matrix([at.get_position() for at in frag])

        # Apply roto-translation
        if self.t is None:
            res = w.apply_R(self.R, pos)
        elif self.R is None:
            res = w.apply_t(self.t, pos)
        else:
            res = w.apply_Rt(self.R, self.t, pos)

        # Copy back to the fragment
        newfrag = deepcopy(frag)
        for i in range(0, len(newfrag)):
            newfrag[i].set_position([float(x) for x in array(res[i, :])[0]])

        return newfrag

    def invert(self):
        """
        Computes the inverse rototranslation.
        """
        self.t = -self.t
        if self.R is not None:
            self.R = self.R.T


class Translation(RotoTranslation):
    """
    This class defines a simple translation.

    Args:
      t (list): the vector describing the translation.
    """
    def __init__(self, t):
        import numpy
        self.R = None
        self.t = numpy.mat(t).reshape(3, 1)
        self.J = 0.0


class Rotation(RotoTranslation):
    """
    This class defines a simple rotation.

    Args:
      t (list): the vector describing the rotation.
    """
    def __init__(self, R):
        self.t = None
        self.R = R
        self.J = 0.0


def interpolate_fragments(A, B, steps, extrapolation_steps=0):
    """
    Given two fragments A and B, this generates a list of Fragments
    that interpolate between A and B in a specified number of steps.

    Args:
      A (BigDFT.Fragments.Fragment): starting fragment.
      B (BigDFT.Fragments.Fragment): ending fragment.
      steps (int): the number of steps to take between A and B.
      extrapolation_steps (int): optionally, we can extrapolate a number of
        steps beyond B on the same trajectory.

    Returns:
      (list): a list of fragments interpolating between A and B including
      A and B.
    """
    from numpy import matrix, array
    from BigDFT.wahba import interpolate_points
    from copy import deepcopy

    pos1 = matrix([at.get_position() for at in A])
    pos2 = matrix([at.get_position() for at in B])

    point_list = interpolate_points(pos1, pos2, steps, extrapolation_steps)

    frag_list = []
    for i in range(0, steps+extrapolation_steps+2):
        new_frag = deepcopy(A)
        for j in range(0, len(new_frag)):
            new_frag[j].set_position([float(x)
                                      for x in array(point_list[i][j])[0]])
        frag_list.append(new_frag)

    return frag_list


class Fragment(MutableSequence):
    """
    A fragment is a list of atoms in a system. Fragment might have quantities
    associated to it, like its electrostatic multipoles (charge, dipole, etc.)
    and also geometrical information (center of mass, principla axis etc.). A
    Fragment might also be rototranslated and combined with other moieteies to
    form a :class:`BigDFT.Systems.Systems`.

    Args:
      atomlist (list): list of atomic dictionaries defining the fragment
      xyzfile (BigDFT.IO.XYZReader): an XYZ file to read from.
      posinp (dict): the posinp style dictionary from a logfile/input file.
      astruct (dict): a BigDFT atomic structure style dictionary.
      system (BigDFT.Systems.System): a BigDFT system, esssentially this
        reduces many fragments into a single fragment.

    .. todo::
       Define and describe if this API is also suitable for solid-state
       fragments

    """

    def __init__(self, atomlist=None, xyzfile=None, posinp=None, astruct=None,
                 system=None):
        from BigDFT.Atoms import Atom
        self.atoms = []

        if system is not None:
            self._system_to_fragment(system)
            return

        # insert atoms.
        if atomlist:
            for atom in atomlist:
                self.append(Atom(atom))
        elif xyzfile:
            with xyzfile:
                for line in xyzfile:
                    self.append(Atom(line))
        elif posinp:
            units = posinp.get('units', 'angstroem')
            for atom in posinp['positions']:
                self.append(Atom(atom, units=units))
        elif astruct:
            units = astruct.get('units', 'angstroem')
            rshift = astruct.get('Rigid Shift Applied (AU)', [0.0, 0.0, 0.0])
            for atom in astruct["positions"]:
                self.append(Atom(atom, units=units))
            self.translate([-1.0*x for x in rshift])

        # Values
        self.q1 = None
        self.q2 = None
        self.frozen = None
        self.conmat = None

    def __len__(self):
        return len(self.atoms)

    def __delitem__(self, index):
        self.atoms.__delitem__(index)

    def insert(self, index, value):
        from BigDFT.Atoms import Atom
        self.atoms.insert(index, Atom(value))

    def __setitem__(self, index, value):
        from BigDFT.Atoms import Atom
        self.atoms.__setitem__(index, Atom(value))

    def __getitem__(self, index):
        # If they ask for only one atom, then we return it as an atom.
        # but if it's a range we return a ``Fragment`` with those atoms in it.
        if isinstance(index, slice):
            return Fragment(atomlist=self.atoms.__getitem__(index))
        else:
            return self.atoms.__getitem__(index)

    def __add__(self, other):
        from copy import deepcopy
        rval = deepcopy(self)
        rval += other
        return rval

    def __radd__(self, other):
        return self if other == 0 else self.__add__(other)

    def __eq__(self, other):
        """
        Compare two fragments. They are equal if all the atoms of the fragments
        are identical.

        other (Fragment): the fragment to compare with.
        """
        ok = True
        for at1, at2 in zip(self, other):
            ok = ok and at1 == at2
            if not ok:
                break
        return ok

    def serialize(self, name, units='bohr'):
        """
        Transform the fragment in a list that can be employed for
        the construction of dataframes or pandas series.

        Args:
            name (str): the name of the fragment
            units (str): the units for the positions
        Returns:
            list: the serialized fragment
        """
        positions = []
        for at in self:
            atdict = {'frag': name}
            atdict.update(at.serialize(units=units))
            positions.append(atdict)
        return positions

    @property
    def centroid(self):
        """
        The center of a fragment.
        """
        from numpy import mean, ravel
        pos = [at.get_position() for at in self]
        return ravel(mean(pos, axis=0))

    def center_of_charge(self):  # , zion):
        """
        The charge center which depends both on the position and net charge
        of each atom.
        """
        from numpy import array
        cc = array([0.0, 0.0, 0.0])
        qtot = 0.0
        for at in self:
            # netcharge = at.q0
            # zcharge = zion[at.sym]
            elcharge = at.nel  # zcharge - netcharge
            cc += elcharge * array(at.get_position())
            qtot += elcharge
        return cc / qtot

    @property
    def q0(self):
        """
        Provides the global monopole of the fragments given as a sum of the
        monopoles of the atoms.
        """
        if len(self) == 0:
            return None
        return [sum(filter(None, [at.q0 for at in self]))]

    def d0(self, center=None):
        """
        Fragment dipole, calculated only from the atomic charges.

        Args:
          center (list): the center of charge of the fragment.
            If this is not present, the centroid is used.
        """
        from numpy import zeros, array
        # one might added a treatment for non-neutral fragments
        # but if the center of charge is used the d0 value is zero
        if center is not None:
            cxyz = center
        else:
            cxyz = self.center_of_charge()  # centroid

        d0 = zeros(3)
        found = False
        for at in self:
            if at.q0 is None:
                found = False
                break
            found = True
            d0 += at.q0 * (array(at.get_position()) - array(cxyz))
        if found:
            return d0
        else:
            return None

    def d1(self, center=None):
        """
        Fragment dipole including the atomic dipoles.

        Args:
          center (list): the center of charge of the fragment.
            If this is not present, the centroid is used.
        """
        from numpy import zeros
        d1 = zeros(3)
        dtot = self.d0(center)
        if dtot is None:
            return dtot

        found = False
        for at in self:
            if at.q1 is not None:
                found = True
                d1 += at.q1

        if found:
            return d1 + dtot
        else:
            return None
        pass

    def ellipsoid(self, center=0.0):
        """
        Todo: define the ellipsoid.
        """
        import numpy as np
        Imat = np.mat(np.zeros(9).reshape(3, 3))
        for at in self:
            rxyz = np.array(at.get_position()) - center
            Imat[0, 0] += rxyz[0]**2  # rxyz[1]**2+rxyz[2]**2
            Imat[1, 1] += rxyz[1]**2  # rxyz[0]**2+rxyz[2]**2
            Imat[2, 2] += rxyz[2]**2  # rxyz[1]**2+rxyz[0]**2
            Imat[0, 1] += rxyz[1] * rxyz[0]
            Imat[1, 0] += rxyz[1] * rxyz[0]
            Imat[0, 2] += rxyz[2] * rxyz[0]
            Imat[2, 0] += rxyz[2] * rxyz[0]
            Imat[1, 2] += rxyz[2] * rxyz[1]
            Imat[2, 1] += rxyz[2] * rxyz[1]
        return Imat

    def get_external_potential(self, units="bohr", charge_offset=False):
        """
        Transform the fragment information into a dictionary ready to be
        put as an external potential.

        Args:
          units (str): the units of the external potential.

        Returns:
          (dict): a dictionary describing the external potential for use in
          an input file.
        """
        pot = [at.get_external_potential(units) for at in self]

        return pot

    def get_net_force(self):
        """
        Returns the net force on a fragment in Ha/Bohr.

        Returns:
          (list) Three values which describe the net force.
        """
        from numpy import array
        ret_val = array([0.0, 0.0, 0.0])
        for at in self:
            ret_val += array(at.get_force())
        return [float(x) for x in ret_val]

    @property
    def qcharge(self):
        """
        The net charge on a fragment.
        """
        netcharge = self.q0[0] if self.q0 is not None else 0
        for at in self:
            netcharge += at.nel
        return netcharge

    @property
    def nel(self):
        """
        The number of valence electrons of the atoms of the fragment
        """
        return sum([at.nel for at in self])

    def rotate_on_axis(self, angle, axis, units="radians"):
        """
        Rotate a fragment along a specific axis.

        Args:
          angle (float): angle to rotate along the axis.
          axis (list): a list of floats defining the vector to rotate along.
          units (str): either radians or degrees.
        """
        from math import pi, cos, sin
        from numpy import mat, array
        from numpy.linalg import norm
        from copy import deepcopy

        # Deal with the units.
        if units == "degrees":
            angle_value = angle * pi/180
        elif units == "radians":
            angle_value = angle
        else:
            raise ValueError("Units must be degrees or radians")

        # Translate back to the origin.
        centroid = self.centroid
        self.translate(-1.0 * centroid)

        # Normalize the unit vector.
        vec = array(axis) / norm(axis)

        # Build the rotation
        c = cos(angle_value)
        s = sin(angle_value)
        t = 1 - c
        x = vec[0]
        y = vec[1]
        z = vec[2]

        rot = mat([
                   [t*x*x + c,    t*x*y - z*s, t*x*z + y*s],
                   [t*x*y + z*s,  t*y*y + c,   t*y*z - x*s],
                   [t*x*z - y*s,  t*y*z + x*s, t*z*z + c]
                  ])

        # Rotate
        rt = Rotation(rot)
        rself = rt.dot(self)
        for i in range(0, len(self)):
            self[i] = deepcopy(rself[i])

        # Translate back
        self.translate(centroid)

    def rotate(self, x=None, y=None, z=None, units="radians"):
        """
        Rotate the fragment.

        Args:
          x (float): angle to rotate on the x axis.
          y (float): angle to rotate on the y axis.
          z (float): angle to rotate on the z axis.
          units (str): either radians or degrees.
        """
        from math import cos, sin, pi
        from numpy import mat, identity
        from copy import deepcopy

        # Deal with the units.
        if units == "degrees":
            if x:
                xval = x * pi / 180
            if y:
                yval = y * pi / 180
            if z:
                zval = z * pi / 180
        elif units == "radians":
            xval = x
            yval = y
            zval = z
        else:
            raise ValueError("Units must be degrees or radians")

        # Translate back to the origin.
        centroid = self.centroid
        self.translate(-1.0 * centroid)

        # Build the rotation
        rot = identity(3)
        if x:
            rx = mat([
                [1.0,        0.0,       0.0],
                [0.0,        cos(xval), -sin(xval)],
                [0.0,        sin(xval),  cos(xval)]
            ])
            rot = rx.dot(rot)
        if y:
            ry = mat([
                [cos(yval), 0.0,        sin(yval)],
                [0.0,        1.0,        0.0],
                [-sin(yval), 0.0,        cos(yval)]
            ])
            rot = ry.dot(rot)
        if z:
            rz = mat([
                [cos(zval), -sin(zval), 0.0],
                [sin(zval),  cos(zval), 0.0],
                [0.0,        0.0,       1.0]
            ])
            rot = rz.dot(rot)

        # Rotate
        rt = Rotation(rot)
        rself = rt.dot(self)
        for i in range(0, len(self)):
            self[i] = deepcopy(rself[i])

        # Translate back
        self.translate(centroid)

    def translate(self, vec):
        """
        Translate the fragment along the vector provided.

        Args:
          vec (list): a list of x, y, z values describing the translation (AU).
        """
        from copy import deepcopy
        rt = Translation(vec)

        trans = rt.dot(self)
        for i in range(0, len(self)):
            self[i] = deepcopy(trans[i])

    def _system_to_fragment(self, sys):
        # Convert to one big system.
        self.__init__()
        for frag in sys.values():
            for at in frag:
                self += [at]

        # Compute the connectivity, first the lookup table.
        if sys.conmat is not None:
            counter = 0
            lookup = {}
            for fragid, frag in sys.items():
                for i, at in enumerate(frag):
                    lookup[(fragid, i)] = counter
                    counter = counter + 1

            # Now the actual connectivity matrix.
            self.conmat = {}
            for fragid, frag in sys.items():
                for i in range(0, len(frag)):
                    t1 = (fragid, i)
                    t2list = sys.conmat[fragid][i]
                    self.conmat[lookup[t1]] = {}
                    for t2, bo in t2list.items():
                        self.conmat[lookup[t1]][lookup[t2]] = bo


def _example():
    """Example of using fragments"""
    from BigDFT.IO import XYZReader, XYZWriter
    from copy import deepcopy

    safe_print("Read in an xyz file and build from a list.")
    atom_list = []
    with XYZReader("SiO") as reader:
        for at in reader:
            atom_list.append(at)
    frag1 = Fragment(atomlist=atom_list)
    for at in frag1:
        safe_print(at.sym, at.get_position())
    safe_print("Centroid", frag1.centroid)
    safe_print()

    safe_print("Build from an xyz file directly.")
    reader = XYZReader("Si4")
    frag2 = Fragment(xyzfile=reader)
    for at in frag2:
        safe_print(at.sym, at.get_position())
    safe_print()

    safe_print("We can combine two fragments with +=")
    frag3 = deepcopy(frag1)
    frag3 += frag2
    for at in frag3:
        safe_print(at.sym, at.get_position())
    safe_print("Length of frag3", len(frag3))
    safe_print()

    safe_print("Since we can iterate easily, we can also write easily.")
    with XYZWriter("test.xyz", len(frag3), "angstroem") as writer:
        for at in frag3:
            writer.write(at)
    with open("test.xyz") as ifile:
        for line in ifile:
            safe_print(line)
    safe_print()

    safe_print("We can also extract using the indices")
    safe_print(dict(frag3[0]))
    sub_frag = frag3[1:3]
    for at in sub_frag:
        safe_print(dict(at))
    safe_print()


if __name__ == "__main__":
    _example()



"""
This module defines some tools used for reading/writing fragment files.

"""
from futile.Utils import write as safe_print

# Standard residue names
# https://cdn.rcsb.org/wwpdb/docs/documentation/file-format/PDB_format_1992.pdf
_standard_residues = ["ALA", "ACD", "ALA", "ALB", "ALI", "ARG", "ARO", "ASN",
                      "ASP", "ASX", "BAS", "CYS", "GLN", "GLU", "GLX", "GLY",
                      "HIS", "HYP", "ILE", "LEU", "LYS", "MET", "PCA", "PHE",
                      "PRO", "SAR", "SER", "THR", "TRP", "TYR", "VAL"]


def _split_line(line, limit):
    split = line.split()
    # print line, split
    if len(split) < 3:  # there should be at least two atoms
        tmp = line.lstrip('CONECT')
        split = ['CONECT'] + _split_malformed_connection(tmp)
    resplit = [split[0]]
    for at2 in split[1:]:
        atom = int(at2)
        if atom > limit:
            atoms = _split_malformed_connection(at2)
        else:
            atoms = [atom]
        resplit += atoms
    return resplit


def _split_malformed_connection(conect):
    maxl = len(conect)
    if maxl > 10:  # we have three atoms at least
        ll = 5 if conect[0] == '9' else 6
        return [conect[:ll]] + _split_malformed_connection(conect[ll:])
    else:
        ll = int(maxl/2)
        tmp1 = conect[:ll]
        tmp2 = conect[ll:]
        return [tmp1, tmp2]


def read_pdb(ifile, include_chain=False, disable_warnings=False):
    """
    Read a system from a PDB file.

    Args:
      ifile (TextIOBase): the file to read from.
      disable_warnings (bool): whether to print warnings about possible file
        issues.
      include_chain (bool): include the chain id if True

    Warning:
       This will read in the connectivity information from the pdb as well.
       However, a pdb file does not provide any information about the bond
       order. Thus, the bond order of each bond will be set to one.

    Returns:
      (BigDFT.Systems.System): the system file.
    """
    from BigDFT.Fragments import Fragment
    from BigDFT.Systems import System
    from warnings import warn
    from BigDFT.UnitCells import UnitCell

    # First pass read in the atoms.
    sys = System()
    lookup = {}
    sys.conmat = {}
    found = False

    for line in ifile:
        try:  # See if there is an atom on this line
            if line[:4] == "ATOM" or line[:6] == "HETATM":
                at, atid, fragid = _process_atom(line,
                                                 include_chain=include_chain)

                # We can ignore lone pairs
                if at.sym == "Lp":
                    continue

                # Add to the system
                if fragid not in sys:
                    sys[fragid] = Fragment()
                sys[fragid] += [at]

                # Build the lookup table
                lookup[atid] = (fragid, len(sys[fragid]) - 1)

            elif line[:6] == "CONECT":
                found = True
                split = _split_line(line, len(lookup))
                (fragid, atnum) = lookup[int(split[1])]

                for at2 in split[2:]:
                    fragid2, atnum2 = lookup[int(at2)]

                    if fragid not in sys.conmat:
                        sys.conmat[fragid] = []
                        for i in range(0, len(sys[fragid])):
                            sys.conmat[fragid].append({})

                    sys.conmat[fragid][atnum][(fragid2, atnum2)] = 1.0

            elif line[:6] == "CRYST1":
                a = float(line[7:15])
                b = float(line[16:24])
                c = float(line[25:33])
                alpha = float(line[34:40])
                beta = float(line[41:47])
                gamma = float(line[48:54])

                sys.cell = UnitCell([a, b, c], units="angstroem")

                if not disable_warnings:
                    if (alpha != 90 or beta != 90 or gamma != 90):
                        warn("Cell angles must be 90 degrees", UserWarning)

        except IndexError:  # For shorter lines
            continue

    if not found:
        sys.conmat = None
    else:
        # for any connectivity not specified we give default values.
        for fragid in sys:
            if fragid not in sys.conmat:
                sys.conmat[fragid] = []
                for i in range(0, len(sys[fragid])):
                    sys.conmat[fragid].append({})

    if not disable_warnings:
        if sum([len(x) for x in sys.values()]) == 0:
            warn("Warning: zero atoms found", UserWarning)

    return sys


def reorder_fragments(frags):
    chs = {}
    for frag in frags:
        chres, num = frag.split(':')
        if "-" in chres:
            ch, res = chres.split('-')
        else:
            ch = 'A'
        subch = chs.setdefault(ch, {})
        if int(num) in subch:
            return frags  # reordering not possible - incoherent fragments
        else:
            subch[int(num)] = frag
    lis = []
    for ch in sorted(chs.keys()):
        for fr in sorted(chs[ch].keys()):
            lis.append(chs[ch][fr])
    return lis


def write_pdb(system, ofile):
    """
    Write out a system to a string in the pdb format.

    Args:
      system (BigDFT.Systems.System): the system to write.
      ofile (TextIOBase): the output stream to write to.
    """

    from BigDFT.Systems import GetFragTuple

    outstr = ""

    # Write Cell
    if system.cell.get_boundary_condition() != "free":
        line = list(" " * 80)
        line[0:6] = "CRYST1".ljust(6)  # CRYST1
        pos = ["{:.3f}".format(x) for x in
               system.cell.get_posinp(units="angstroem")]
        line[7:15] = pos[0].rjust(8)  # a (Angstrom)
        line[16:24] = pos[1].rjust(8)  # b (Angstrom)
        line[25:33] = pos[2].rjust(8)  # c (Angstrom)
        line[34:40] = "90".rjust(6)  # alpha (degrees)
        line[41:47] = "90".rjust(6)  # beta (degrees)
        line[48:54] = "90".rjust(6)  # gamma (degrees)
        outstr += "".join(line) + "\n"

    # Write Postions
    idx = 1
    lookup = {}
    reorder = reorder_fragments(system)
    for fragid in reorder:  # system.items():
        frag = system[fragid]
        if any([x in fragid for x in _standard_residues]):
            atom_type = "ATOM".ljust(6)
        else:
            atom_type = "HETATM"
        for i, at in enumerate(frag):
            pos = [str("{:.3f}".format(x))
                   for x in at.get_position("angstroem", cell=system.cell)]
            fragtuple = GetFragTuple(fragid)
            resname = fragtuple[0]
            chres = resname.split('-')
            chain = 'A' if '-' not in resname else chres[0]
            resname = chres[-1][:3].ljust(3)
            line = list(" " * 80)
            line[0:6] = atom_type  # HETATM
            line[6:11] = str(idx).rjust(5)  # SERIAL NUMBER
            if "name" in at:  # ATOM NAME
                line[12:16] = at["name"].ljust(4)
            else:
                line[12:16] = at.sym.ljust(4)
            line[16:17] = " "  # ALTERNATIVE LOCATION INDICATOR
            line[17:20] = resname  # RESIDUE NAME
            line[21:22] = chain  # CHAIN IDENTIFIER
            line[22:26] = fragtuple[1].rjust(4)  # RESIDUE SEQUENCE NUMBER
            line[26:27] = " "  # CODE FOR INSERTION OF RESIDUES
            line[30:38] = pos[0].rjust(8)  # X COORDINATE
            line[38:46] = pos[1].rjust(8)  # Y COORDINATE
            line[46:54] = pos[2].rjust(8)  # Z COORDINATE
            line[55:61] = " 1.00 "  # OCCUPANCY
            line[61:67] = " 0.00 "  # TEMPERATURE
            line[72:76] = " B  "  # SEGMENT IDENTIFIER
            line[76:78] = at.sym.rjust(2)  # ELEMENT SYMBOL
            line[78:80] = "  "  # CHARGE
            outstr += "".join(line) + "\n"

            # Keep track of the indexes
            lookup[(fragid, i)] = idx
            idx = idx + 1

    # Write the connectivity information
    if system.conmat is not None:
        for fragid, frag in system.items():
            for i, at in enumerate(frag):
                connections = system.conmat[fragid][i]
                connections = [lookup[x] for x in connections.keys()]

                line = list(" " * 80)
                line[0:6] = "CONECT"

                # SERIAL NUMBER
                line[7:11] = str(lookup[(fragid, i)]).rjust(4)
                if len(connections) > 0:  # BOND SERIAL NUMBERS
                    line[12:16] = str(connections[0]).rjust(4)
                if len(connections) > 1:
                    line[17:21] = str(connections[1]).rjust(4)
                if len(connections) > 2:
                    line[22:26] = str(connections[2]).rjust(4)
                if len(connections) > 3:
                    line[27:31] = str(connections[3]).rjust(4)

                outstr += "".join(line) + "\n"

    ofile.write(outstr)


def read_mol2(ifile, disable_warnings=False):
    """
    Read a system from a mol2 file.

    Args:
      ifile (TextIOBase): the file to read from.
      disable_warnings (bool): whether to print warnings about possible file
        issues.

    Returns:
      (BigDFT.Systems.System): the system file.
    """
    from BigDFT.Systems import System
    from BigDFT.Fragments import Fragment
    from BigDFT.Atoms import Atom
    from BigDFT.UnitCells import UnitCell
    from warnings import warn

    sys = System()

    # Just go ahead and read the whole file into a string.
    lines = [x for x in ifile]

    # First pass, read in the number of atoms.
    for start, line in enumerate(lines):
        if ("@<TRIPOS>MOLECULE" in line):
            break
    start += 1

    split = lines[start+1].split()
    natoms = int(split[0])
    nbonds = int(split[1])

    # Second pass read in the atoms.
    for start, line in enumerate(lines):
        if ("@<TRIPOS>ATOM" in line):
            break
    start += 1

    lookup = []
    for i in range(0, natoms):
        split = lines[start + i].split()
        pos = [float(x) for x in split[2:5]]
        name = split[5]
        sym = name.split(".")[0]
        fragid = split[7] + ":" + split[6]
        charge = [float(split[8])]

        # Add fragment
        if fragid not in sys:
            sys[fragid] = Fragment()
        at = Atom({sym: pos, "units": "angstroem", "q0":
                   charge, "name": name})
        sys[fragid] += [at]

        # Lookup table for connectivity
        lookup.append((fragid, len(sys[fragid]) - 1))

    # Third pass reads the connectivity.
    for start, line in enumerate(lines):
        if ("@<TRIPOS>BOND" in line):
            break
    start += 1

    if start < len(lines):
        sys.conmat = {}
        for fragid, frag in sys.items():
            sys.conmat[fragid] = []
            for i in range(0, len(frag)):
                sys.conmat[fragid].append({})

    bowarn = False
    for i in range(0, nbonds):
        split = lines[start + i].split()
        frag1, at1 = lookup[int(split[1])-1]
        frag2, at2 = lookup[int(split[2])-1]
        bo = split[3]
        try:
            bo = float(split[3])
        except ValueError:
            bowarn = True
            bo = 1
        sys.conmat[frag1][at1][(frag2, at2)] = bo

        # Since mol2 doesn't include the symmetric bonds.
        if frag1 != frag2 or at1 != at2:
            sys.conmat[frag2][at2][(frag1, at1)] = bo

    # Fourth path reads the unit cell.
    for start, line in enumerate(lines):
        if ("@<TRIPOS>CRYSIN" in line):
            break
    start += 1

    if start < len(lines):
        split = lines[start].split()
        a = float(split[0])
        b = float(split[1])
        c = float(split[2])
        alpha = float(split[3])
        beta = float(split[4])
        gamma = float(split[5])

        sys.cell = UnitCell([a, b, c], units="angstroem")

        if not disable_warnings:
            if (alpha != 90 or beta != 90 or gamma != 90):
                warn("Cell angles must be 90 degrees", UserWarning)

    if not disable_warnings:
        if sum([len(x) for x in sys.values()]) == 0:
            warn("Warning: zero atoms found", UserWarning)
        if bowarn:
            warn("Unsupported bond type had to be set to 1 (i.e. aromatic)",
                 UserWarning)

    return sys


def write_mol2(system, ofile):
    """
    Write out a system to a string in the mol2 format.

    Args:
      system (BigDFT.Systems.System): the system to write.
      ofile (TextIOBase): the output stream to write to.
    Returns:
      (str): a string representation of the file.
    """
    from BigDFT.Systems import GetFragTuple

    charges = True

    idx = 1
    lookup = {}
    atomstr = "@<TRIPOS>ATOM\n"
    for fragid, frag in system.items():
        for i, at in enumerate(frag):
            pos = [str("{:.3f}".format(x))
                   for x in at.get_position("angstroem", cell=system.cell)]
            fragtuple = GetFragTuple(fragid)

            line = str(idx) + " " + at.sym+str(idx) + " "
            line += pos[0] + " " + pos[1] + " " + pos[2] + " "
            line += at.sym + " "
            line += fragtuple[1] + " "
            line += fragtuple[0] + " "

            if at.q0:
                line += str(at.q0)
            else:
                line += str(0.0)

            atomstr += line + "\n"

            # Keep track of the indexes
            lookup[(fragid, i)] = idx
            idx = idx + 1
    numatoms = idx - 1

    # Write the connectivity information
    bondstr = "@<TRIPOS>BOND\n"
    numbonds = 0
    idx = 1
    if system.conmat is not None:
        for fragid, frag in system.items():
            for i, at in enumerate(frag):
                connections = system.conmat[fragid][i]
                bonds = connections.values()
                connections = [lookup[x] for x in connections.keys()]

                for con, bo in zip(connections, bonds):
                    if con <= idx:
                        continue
                    bondstr += str(numbonds + 1) + " "
                    bondstr += str(idx) + " " + str(con) + " "
                    bondstr += str(int(bo)) + "\n"
                    numbonds = numbonds + 1
                idx = idx + 1

    # Write out the molecule part
    molstr = "@<TRIPOS>MOLECULE\n"
    molstr += "generated by BigDFT\n"
    molstr += str(numatoms) + " " + str(numbonds) + " "
    molstr += str(len(system.keys())) + " 0 0\n"
    molstr += "****\n"
    if charges:
        molstr += "MULLIKEN_CHARGES\n"
    else:
        molstr += "NO_CHARGES\n"
    molstr += "****\n"
    molstr += "****\n"

    # Write Cell
    cellstr = ""
    if system.cell.get_boundary_condition() != "free":
        cellstr += "@<TRIPOS>CRYSIN\n"
        pos = ["{:.3f}".format(x) for x in
               system.cell.get_posinp(units="angstroem")]
        cellstr += " ".join(pos)
        cellstr += " 90 90 90"
        cellstr += " 1 1\n"

    ofile.write(molstr + "\n" + atomstr + "\n" + bondstr + "\n" + cellstr)


def write_xyz(system, ofile):
    """
    Write out a system to a file in the xyz format.

    Args:
      system (BigDFT.Systems.System): the system to write.
      ofile (TextIOBase): the output stream to write to.
      cell (list): the unit cell.
    """
    outstr = ""
    outstr += str(sum([len(x) for x in system.values()])) + " angstroem\n"
    outstr += system.cell.get_boundary_condition("angstroem") + "\n"

    for frag in system.values():
        for at in frag:
            pos = at.get_position("angstroem", cell=system.cell)
            outstr += at.sym + " " + " ".join([str(x) for x in pos]) + "\n"

    ofile.write(outstr)


def _process_atom(line, include_chain=False):
    """
    This processes a line of a pdb file and extracts information about an atom.

    Returns:
      (BigDFT.Atoms.Atom, int, str): return the Atom on this line, its id in
      the pdb, and the fragment it belongs to. It may include the chain id.
    """
    from BigDFT.Atoms import Atom
    # Get the basic information about this atom
    sym = line[76:78].strip().capitalize()

    if "1-" in sym:
        sym = sym.replace("1-", "")
    if "1+" in sym:
        sym = sym.replace("1+", "")

    fullname = line[12:16]

    name = fullname.strip()

    xpos = float(line[30:38])
    ypos = float(line[38:46])
    zpos = float(line[46:54])

    at = Atom({"sym": sym, "r": [xpos, ypos, zpos], "name": name,
               "units": "angstroem"})

    # Get the atom id for building the lookup table
    atid = int(line[6:11])

    # Information about the residue
    resname = line[17:20]
    chain = line[20:22].strip()
    resid = str(int(line[22:26]))
    fragid = resname+":"+resid
    if include_chain:
        fragid = chain+'-'+fragid

    return at, atid, fragid


class XYZReader():
    """
    A class which can be used to read from xyz files.

    This class should behave like a standard file, which means you can use
    it in ``with`` statements, and use the ``next`` command to iterate over it.

    Attributes:
      closed (bool): check if a file is open or not.
      units (str): the units that the xyz file was in.
      natoms (int): the number of atoms in the xyz file.
      cell (list): a list of floats describing the cell.

    Args:
      filename (str): the file to read from. You can also specify a molecule
        that might be in the database.
    """

    def __init__(self, filename):
        self.filename = filename
        self.natoms = None
        self.cell = None
        self.atoms_positions = []

    def open(self):
        self.__enter__()

    def __enter__(self):
        from os.path import join, abspath, dirname
        from BigDFT.UnitCells import UnitCell

        try:
            self._handle = open(self.filename, "r")
        except IOError as e:  # see if the file is in the database
            dirXYZ = join(dirname(__file__), "Database", "XYZs")
            filename = abspath(join(dirXYZ, self.filename + ".xyz"))
            try:
                self._handle = open(filename, "r")
            except IOError:  # Raise the error from the original path
                raise e

        # Read atoms and units.
        line = next(self._handle)
        self.natoms = int(line.split()[0])
        try:
            self.units = line.split()[1]
        except IndexError:
            self.units = "angstroem"

        # Read boundary condition.
        line = next(self._handle).split()
        if len(line) == 0:
            self.cell = UnitCell()
        elif line[0] == "wire":
            self.cell = UnitCell([float("inf"), float("inf"),
                                  float(line[3])], units=self.units)
        elif line[0] == "surface":
            self.cell = UnitCell([float(line[1]), float("inf"),
                                  float(line[3])], units=self.units)
        elif line[0] == "periodic":
            self.cell = UnitCell([float(x) for x in line[1:4]],
                                 units=self.units)
        else:
            self.cell = UnitCell()

        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        from BigDFT.Atoms import Atom
        line = next(self._handle)
        if self.natoms == len(self.atoms_positions):
            raise StopIteration
        split = line.split()
        sym = split[0]
        position = [float(x) for x in split[1:4]]
        this_pos = Atom({'sym': sym, 'r': position, "units": self.units})
        self.atoms_positions.append(this_pos)
        return this_pos

    def __iter__(self):
        return self

    def close(self):
        if self.natoms != len(self.atoms_positions):
            raise IOError('The number of atoms is not consistent with the'
                          ' number of lines')
        self.__exit__()

    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
        self._handle.close()

    @property
    def closed(self):
        return self._handle.closed


class XYZWriter():
    """
    A class for writing XYZ files.

    This class should behave like a standard file, which means you can use
    it in ``with`` statements and write.

    Args:
      filename (str): the file to write to.
      natoms (int): how many atoms we will write.
      units (str): the units of the file. Defaults to angstroem.
      cell (BigDFT.UnitCells.UnitCell): the unit cell.
    """

    def __init__(self, filename, natoms, units="angstroem", cell=None):
        from BigDFT.UnitCells import UnitCell
        self.filename = filename
        self.natoms = natoms
        self.units = units
        if cell is None:
            self.cell = UnitCell()
        else:
            self.cell = cell

    def open(self):
        self.__enter__()

    def __enter__(self):
        self._handle = open(self.filename, "w")
        self._handle.write(str(self.natoms) + " ")
        self._handle.write(self.units)
        self._handle.write("\n")

        # The unit cell
        self._handle.write(self.cell.get_boundary_condition(self.units))
        self._handle.write("\n")

        return self

    def write(self, atomdict):
        """
        Write an atom to the file.

        Args:
          atom (dict): a dictionary describing the atom.
        """
        from BigDFT.Atoms import Atom
        at = Atom(atomdict)
        self._handle.write(at.sym + " ")
        pos = at.get_position(self.units)
        self._handle.write(" ".join([str(x) for x in pos]))
        self._handle.write("\n")

    def close(self):
        self.__exit__()

    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
        self._handle.close()

    @property
    def closed(self):
        return self._handle.closed


def _example():
    """Test the XYZ Module"""
    from BigDFT.Systems import System
    from BigDFT.Fragments import Fragment
    from BigDFT.UnitCells import UnitCell
    file = "Si4"

    safe_print("First let's try reading an XYZ file.")
    atom_list = []
    with XYZReader(file) as reader:
        safe_print(reader.closed)
        for at in reader:
            atom_list.append(at)
    safe_print(reader.closed)
    safe_print(atom_list)
    safe_print()

    safe_print("Now let's try writing an XYZ file.")
    safe_print()
    with XYZWriter("test.xyz", len(atom_list),
                   units=reader.units) as writer:
        safe_print(writer.closed)
        for at in atom_list:
            writer.write(at)
    safe_print(writer.closed)
    safe_print()
    with open("test.xyz") as ifile:
        for line in ifile:
            safe_print(line, end='')
    safe_print()

    safe_print("Print with various boundary conditions")
    with XYZWriter("test.xyz", len(atom_list), reader.units,
                   cell=UnitCell()) as writer:
        for at in atom_list:
            writer.write(at)
    with XYZReader("test.xyz") as ifile:
        print(ifile.cell.get_boundary_condition())

    with XYZWriter("test.xyz", len(atom_list), reader.units,
                   cell=UnitCell([5, 5, 5])) as writer:
        for at in atom_list:
            writer.write(at)
    with XYZReader("test.xyz") as ifile:
        print(ifile.cell.get_boundary_condition())

    with XYZWriter("test.xyz", len(atom_list), reader.units,
                   cell=UnitCell([5, float("inf"), 5])) as writer:
        for at in atom_list:
            writer.write(at)
    with XYZReader("test.xyz") as ifile:
        print(ifile.cell.get_boundary_condition())

    with XYZWriter("test.xyz", len(atom_list), reader.units,
                   cell=UnitCell([float("inf"), float("inf"), 5])) as writer:
        for at in atom_list:
            writer.write(at)
    with XYZReader("test.xyz") as ifile:
        print(ifile.cell.get_boundary_condition())
    safe_print()

    safe_print("Now let's demonstrate the pdb and mol2 writer")
    sys = System()
    sys["FRAG:0"] = Fragment(atom_list)
    with open("test.pdb", "w") as ofile:
        write_pdb(sys, ofile)
    with open("test.pdb") as ifile:
        for line in ifile:
            safe_print(line, end='')
    safe_print()

    with open("test.mol2", "w") as ofile:
        write_mol2(sys, ofile)
    with open("test.mol2") as ifile:
        for line in ifile:
            safe_print(line, end='')
    safe_print()


if __name__ == "__main__":
    _example()



"""A module to define typical operations that can be done on biological systems
   starting from the API and the PDB standards employed in the Polaris code

"""


def read_polaris_sle(slefile):
    """
    Read the setup file of polaris with the atomic information.

    Args:
        slefile(str): the sle polaris file.

    Returns:
        list: a list of the atoms of the system with the same order of
            the pdbfile, ready to be passed to the atoms attribute
    """
    attributes = []
    iat = 0
    with open(slefile) as ifile:
        for line in ifile:
            sleline = line.split()
            try:
                zion, ch, pn, mass, chg,\
                    polchg, chi, name, chain, iat, ires, res = sleline
                attributes.append({'q0': [float(chg)], 'name': name,
                                   'resnum': ':'.join(map(str, (res, ires)))})
            except Exception:
                pass
    return attributes


def write_bigdft_charges_in_sle(charges, slefile, outfile):
    """
    Override polaris charges coming from BigDFT file in the sle

    Args:
        charges (dict): dictionary of the {iat: charges}, with iat from the pdb
        slefile (str): the sle polaris file.
        outfile (str): the output file
    """
    fout = open(outfile, 'w')
    with open(slefile) as ifile:
        for line in ifile:
            sleline = line.split()
            try:
                zion, ch, pn, mass, chg,\
                    polchg, chi, name, chain, iat, ires, res = sleline
                newchg = charges[int(iat)]
                fout.write(' '.join(map(str, (zion, ch, pn, mass, newchg,
                                              polchg, chi, name, chain, iat,
                                              ires, res, '\n'))))
            except Exception:
                fout.write(line)
    fout.close()


def read_polaris_pdb(pdbfile, chain_as_letter=False, slefile=None):
    """
    Read coordinates in the PDB format of POLARIS

    Args:
       pdbfile (str): path of the input file
       chain_as_letter (bool): If True, the fifth column
           is assumed to contain a letter
       slefile (str): path of the file ``.sle`` of Polaris from which
           to extract the system's attributes.

    Warning:
       Assumes Free Boundary conditions for the molecule.
       Only accepts atoms that have one letter in the symbol.
       Switch representation if there is a single letter in the fifth column

    Returns:
       System: A system class
    """
    from BigDFT.Fragments import Fragment
    from BigDFT.Systems import System, GetFragId
    from BigDFT.Atoms import Atom
    sys = System()
    units = 'angstroem'
    with open(pdbfile) as ifile:
            for line in ifile:
                if 'ATOM' not in line:
                    continue
                atomline = line.split()
                if chain_as_letter:
                    iat, name, frag, lett, ifrag, x, y, z, sn = atomline[1:10]
                    chain = lett
                    segname = sn
                else:
                    iat, name, frag, ifrag, x, y, z, chlett = atomline[1:9]
                    chain = chlett[2]
                    segname = chlett
                atdict = {str(name[:1]): map(float, [x, y, z]),
                          'frag': [chain+'-'+frag, int(ifrag)], 'name': name,
                          'iat': int(iat), 'segname': segname}
                fragid = GetFragId(atdict, iat)
                if fragid not in sys:
                    sys[fragid] = Fragment()
                sys[fragid].append(Atom(atdict, units=units))
    if slefile is None:
        return sys
    attributes = read_polaris_sle(slefile)
    from BigDFT import Systems as S, Fragments as F, Atoms as A
    system = S.System()
    for name, frag in sys.items():
        refrag = F.Fragment()
        for at in frag:
            atdict = at.dict()
            att = attributes[atdict['iat']-1]
            assert att['name'] == atdict['name']
            atdict.update(att)
            refrag.append(A.Atom(atdict))
        system[name] = refrag
    return system


def split_pdb_trajectory(directory, filename, prefix):
    """
    Split a trajectory file into various PDB files containing a single
    snapshot.

    Args:
        directory (str): the path of the directory in which the trajectory
            file exists
        filename (str): the trajectory file name
        prefix (str): the name to be provided to the resulting files.
           Files are numbered by prefix_0,_1, etc.

    Returns:
        list: list of the files created, including the directory
    """
    files_to_treat = []
    from os.path import join
    f = open(join(directory, filename))
    fileid = 0
    for line in f.readlines():
        if 'CRYST1' in line:
            newfilename = join(directory, prefix + str(fileid) + '.pdb')
            newf = open(newfilename, 'w')
        newf.write(line)
        if 'ENDMDL' in line:
            newf.close()
            files_to_treat.append(newfilename)
            fileid += 1
    f.close()
    return files_to_treat


def convert_pdb(filein, fileout, **kwargs):
    """
    Call the `func:read_polaris_pdb` function to convert the system, and verify
    that the converted filename is giving the same system.

    Args:
        filein (str): input file path
        fileout (str): output file path. Can be identical to filein
        **kwargs: arguments to be passed to `func:read_polaris_pdb`

    Returns:
        tuple: the systems read from filein and fileout, respectively
    """
    from BigDFT.IO import read_pdb, write_pdb
    sys1 = read_polaris_pdb(filein, **kwargs)
    write_pdb(sys1, open(fileout, 'w'))
    # Verify
    sys2 = read_pdb(open(fileout))
    return sys1, sys2



"""
This module contains some wrappers for using OpenMM to perform
various operations on BigDFT molecules.

https://simtk.org/api_docs/openmm/api6_0/python/index.html
"""

from BigDFT.Systems import System


def get_available_ff_names():
    import os
    import sys
    names = []
    for path in sys.path:
        for root, dirs, files in os.walk(path):
            for file in files:
                if file.endswith(".xml"):
                    filename = os.path.join(root, file)
                    if 'data' in filename and 'openmm' in filename:
                        iname = filename.index('data')+5
                        names.append(filename[iname:])
    return names


class OMMSystem(System):
    """
    Class of OpenMM binder system which enables functionalities
    of openMM on a BigDFT system

    Args:
        system (System): a instance of a system class
        filename (str): name of the PDB file to instantiate the system
    """
    def __init__(self, system=None, filename=None):
        from simtk.openmm.app.pdbfile import PDBFile
        from simtk.openmm import app
        from BigDFT.IO import read_pdb, write_pdb
        from tempfile import NamedTemporaryFile as tmp
        if filename is not None:
            pdb = PDBFile(open(filename))
            sys = read_pdb(open(filename))
        elif system is not None:
            sys = system
            ofile = tmp('w')
            write_pdb(system=system, ofile=ofile)
            ofilename = ofile.name
            pdb = PDBFile(open(ofilename))
            ofile.close()
        System.__init__(self, **sys.dict())
        self.pdb = pdb
        self.modeller = app.Modeller(pdb.topology, pdb.positions)

    def set_forcefields(self, *ff_list):
        """
        Define a set of force fields that will be used in the

        Args:
            *ff_list: list of the force fields to be included,
                in priority order.
                Use the :func:py:`get_available_ff_names` to identify the
                available force fields
        """
        from simtk.openmm import app
        self.forcefield = app.ForceField(*ff_list)

    @property
    def OMMsystem(self):
        from simtk.openmm import app
        if not hasattr(self, '_system'):
            self._system = self.forcefield.createSystem(
                self.modeller.topology, nonbondedMethod=app.NoCutoff,
                constraints=None)
        return self._system

    def set_integrator(self, T=298.15):
        from simtk import unit as u
        import simtk.openmm as mm
        temperature = T * u.kelvin
        self.integrator = mm.LangevinIntegrator(
            temperature, 1 / u.picosecond,  0.0005 * u.picoseconds)

    @property
    def OMMsimulation(self):
        from simtk.openmm import app
        if not hasattr(self, '_simulation'):
            self._simulation = app.Simulation(self.modeller.topology,
                                              self.OMMsystem, self.integrator)
            self._simulation.context.setPositions(self.modeller.positions)
        return self._simulation

    def OMMenergy(self, units='kcal/mol'):
        from simtk.openmm import KcalPerKJ
        energy = self.OMMsimulation.context.getState(
             getEnergy=True).getPotentialEnergy()
        return energy._value * KcalPerKJ

    @property
    def OMMposition(self):
        return self.OMMsimulation.context.getState(
            getPositions=True).getPositions()

    def write(self, ofile):
        from simtk.openmm import app
        app.PDBFile.writeFile(self.OMMsimulation.topology, self.OMMposition,
                              open(ofile, 'w'))

    def optimize(self, iters):
        self.OMMsimulation.minimizeEnergy(maxIterations=iters)

    @property
    def forcegroups(self):
        forcegroups = {}
        for i in range(self.OMMsystem.getNumForces()):
            force = self.OMMsystem.getForce(i)
            force.setForceGroup(i)
            forcegroups[force] = i
        return forcegroups

    def get_energies(self):
        from simtk.openmm import KcalPerKJ

        def component_name(k):
            return str(k).split('.')[-1].split(';')[0].lstrip('"')

        energies = {}
        for f, i in self.forcegroups.items():
            en = self.OMMsimulation.context.getState(
                getEnergy=True, groups=2**i).getPotentialEnergy()
            name = component_name(f)
            energies[name] = en._value * KcalPerKJ
        return energies
